from .rxdmath import _Arithmeticed
import weakref
from .section1d import Section1D
from neuron import h, nrn, nrn_dll_sym
from . import node, nodelist, rxdmath, region
import numpy
import warnings
import itertools
from . import options
from .rxdException import RxDException
from . import initializer
from collections.abc import Callable


import ctypes
import re

# Now set in rxd.py
# set_nonvint_block = neuron.nrn_dll_sym('set_nonvint_block')

# Update the structure_change_cnt & diam_change_cnt if the shape has changed
_nrn_shape_update = nrn_dll_sym("nrn_shape_update_always")

fptr_prototype = ctypes.CFUNCTYPE(None)

set_setup = nrn_dll_sym("set_setup")
set_setup.argtypes = [fptr_prototype]
set_initialize = nrn_dll_sym("set_initialize")
set_initialize.argtypes = [fptr_prototype]

# setup_solver = nrn.setup_solver
# setup_solver.argtypes = [ctypes.py_object, ctypes.py_object, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double, ctypes.c_double]

ECS_insert = nrn_dll_sym("ECS_insert")
ECS_insert.argtypes = [
    ctypes.c_int,
    ctypes.py_object,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_double,
    ctypes.c_double,
    ctypes.c_double,
    ctypes.c_double,
    ctypes.c_double,
    ctypes.c_double,
    ctypes.py_object,
    ctypes.py_object,
    ctypes.c_int,
    ctypes.c_double,
    ctypes.c_double,
]

ECS_insert.restype = ctypes.c_int

ICS_insert = nrn_dll_sym("ICS_insert")
ICS_insert_inhom = nrn_dll_sym("ICS_insert_inhom")

ICS_insert_inhom.argtypes = ICS_insert.argtypes = [
    ctypes.c_int,
    ctypes.py_object,
    ctypes.c_long,
    numpy.ctypeslib.ndpointer(dtype=int),
    numpy.ctypeslib.ndpointer(dtype=int),
    ctypes.c_long,
    numpy.ctypeslib.ndpointer(dtype=int),
    ctypes.c_long,
    numpy.ctypeslib.ndpointer(dtype=int),
    ctypes.c_long,
    numpy.ctypeslib.ndpointer(dtype=float),
    ctypes.c_double,
    ctypes.c_bool,
    ctypes.c_double,
    numpy.ctypeslib.ndpointer(dtype=float),
]

ICS_insert.restype = ctypes.c_int
ICS_insert_inhom.restype = ctypes.c_int

# function to change extracellular diffusion
set_diffusion = node.set_diffusion

species_atolscale = nrn_dll_sym("species_atolscale")
species_atolscale.argtypes = [
    ctypes.c_int,
    ctypes.c_double,
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_int),
]

remove_species_atolscale = nrn_dll_sym("remove_species_atolscale")
remove_species_atolscale.argtypes = [ctypes.c_int]

_set_grid_concentrations = nrn_dll_sym("set_grid_concentrations")
_set_grid_concentrations.argtypes = [
    ctypes.c_int,
    ctypes.c_int,
    ctypes.py_object,
    ctypes.py_object,
]

_ics_set_grid_concentrations = nrn_dll_sym("ics_set_grid_concentrations")
_ics_set_grid_concentrations.argtypes = [
    ctypes.c_int,
    ctypes.c_int,
    numpy.ctypeslib.ndpointer(dtype=numpy.int64),
    numpy.ctypeslib.ndpointer(dtype=numpy.int64),
    ctypes.py_object,
]

_set_grid_currents = nrn_dll_sym("set_grid_currents")
_set_grid_currents.argtypes = [
    ctypes.c_int,
    ctypes.c_int,
    ctypes.py_object,
    ctypes.py_object,
    ctypes.py_object,
]

_ics_set_grid_currents = nrn_dll_sym("ics_set_grid_currents")
_ics_set_grid_currents.argtypes = [
    ctypes.c_int,
    ctypes.c_int,
    ctypes.py_object,
    numpy.ctypeslib.ndpointer(dtype=float),
]


_delete_by_id = nrn_dll_sym("delete_by_id")
_delete_by_id.argtypes = [ctypes.c_int]

# function to change extracellular tortuosity
_set_tortuosity = nrn_dll_sym("set_tortuosity")
_set_tortuosity.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.py_object]


# The difference here is that defined species only exists after rxd initialization
_all_species = []
_defined_species = {}
_all_defined_species = []
_defined_species_by_gid = []


def _get_all_species():
    return _all_defined_species


_extracellular_diffusion_objects = weakref.WeakKeyDictionary()
_intracellular_diffusion_objects = weakref.WeakKeyDictionary()

_species_count = 0

_has_1d = False
_has_3d = False


def _update_tortuosity(region):
    """Update tortuosity for all species on region"""

    for s, r in _extracellular_diffusion_objects.items():
        if (
            r == region
            and hasattr(s, "_id")
            and not hasattr(s, "_deleted")
            and not s._diffusion_characteristic
        ):
            _set_tortuosity(s._id, region._permeability_vector)


def _update_volume_fraction(region):
    """Update volume fractions for all species on region"""

    for s, r in _extracellular_diffusion_objects.items():
        if (
            r == region
            and hasattr(s, "_id")
            and not hasattr(s, "_deleted")
            and not s._diffusion_characteristic
        ):
            _set_volume_fraction(s._id, region._volume_faction_vector)


def _1d_submatrix_n():
    if not _has_1d:
        return 0
    else:
        return len(node._states)
    """elif not _has_3d:
        return len(node._states)
    else:
        return numpy.min([sp()._indices3d() for sp in list(_get_all_species()) if sp() is not None])"""


_ontology_id = re.compile(
    "^\[[a-zA-Z][a-zA-Z0-9_]*:[a-zA-Z0-9_]*\]|[a-zA-Z][a-zA-Z0-9_]*:[a-zA-Z0-9_]*$"
)


class _SpeciesMathable(object):
    # support arithmeticing
    def __neg__(self):
        return -1 * _Arithmeticed(self)

    def __Pos__(self):
        return _Arithmeticed(self)

    def __pow__(self, other):
        return _Arithmeticed(self) ** other

    def __add__(self, other):
        return _Arithmeticed(self) + other

    def __sub__(self, other):
        return _Arithmeticed(self) - other

    def __mul__(self, other):
        return _Arithmeticed(self) * other

    def __div__(self, other):
        return _Arithmeticed(self) / other

    def __rdiv__(self, other):
        return _Arithmeticed(other) / self

    def __radd__(self, other):
        return _Arithmeticed(other) + self

    def __rmul__(self, other):
        return _Arithmeticed(self) * other

    def __truediv__(self, other):
        return _Arithmeticed(self) / other

    def __rtruediv__(self, other):
        return _Arithmeticed(other) / self

    def __rfloordiv__(self, other):
        return _Arithmeticed(other) / self

    def __rsub__(self, other):
        return _Arithmeticed(other) - self

    def __abs__(self):
        return abs(_Arithmeticed(self))

    def _evaluate(self, location):
        return _Arithmeticed(self)._evaluate(location)

    def __ne__(self, other):
        other = rxdmath._ensure_arithmeticed(other)
        self2 = rxdmath._ensure_arithmeticed(self)
        rxdmath._validate_reaction_terms(self2, other)
        return rxdmath._Reaction(self2, other, "<>")

    def __gt__(self, other):
        other = rxdmath._ensure_arithmeticed(other)
        self2 = rxdmath._ensure_arithmeticed(self)
        rxdmath._validate_reaction_terms(self2, other)
        return rxdmath._Reaction(self2, other, ">")

    def __lt__(self, other):
        other = rxdmath._ensure_arithmeticed(other)
        self2 = rxdmath._ensure_arithmeticed(self)
        rxdmath._validate_reaction_terms(self2, other)
        return rxdmath._Reaction(self2, other, "<")

    def _semi_compile(self, reg, instruction):
        from . import region

        # region is Extracellular

        # if the species is only define on one extracellular region then use that
        if len(self._regions) == 0 and len(self._extracellular_regions) == 1:
            reg = self._extracellular_regions[0]

        # region is Extracellular
        if isinstance(reg, region.Extracellular) or len(self._regions) == 0:
            ecs_instance = self._extracellular_instances[reg]
            return ecs_instance._semi_compile(reg, instruction)
        # region is 3d intracellular
        if isinstance(reg, region.Region) and instruction == "do_3d":
            ics_instance = self._intracellular_instances[reg]
            return ics_instance._semi_compile(reg, instruction)
        if isinstance(reg, region.Region) and instruction == "do_1d":
            if reg in self._regions:
                return "species[%d][%d]" % (self._id, reg._id)
            else:
                raise RxDException(
                    "Species %r is not defined on region %r." % (self, reg)
                )
        raise RxDException("_semi_compile received inconsistent state")

    def _involved_species(self, the_dict):
        the_dict[self._semi_compile] = weakref.ref(self)

    @property
    def d(self):
        return self._d

    @d.setter
    def d(self, value):
        from . import rxd

        self._d = value
        if hasattr(self, "_extracellular_instances"):
            for ecs in self._extracellular_instances.values():
                ecs.d = value
        if hasattr(self, "_intracellular_instances"):
            for ics in self._intracellular_instances.values():
                ics.d = value
        if initializer.is_initialized() and hasattr(self, "_region_indices"):
            _volumes, _surface_area, _diffs = node._get_data()
            if _diffs.size > 0:
                if hasattr(value, "__len__"):
                    raise RxDException(
                        "Intracellular 1D can only have one diffusion coefficient. To set 3D intracellular or extracellular diffusion, access the species on the regions; `species[region].d = (Dx, Dy, Dz)`"
                    )
                _diffs[self._indices1d()] = value
                rxd._setup_matrices()


class SpeciesOnExtracellular(_SpeciesMathable):
    def __init__(self, species, extracellular):
        """The restriction of a Species to a Region."""
        self._species = weakref.ref(species)
        self._extracellular = weakref.ref(extracellular)
        self._id = _species_count

    def __repr__(self):
        return "%r[%r]" % (self._species(), self._extracellular()._region)

    def _short_repr(self):
        return "%s[%s]" % (
            self._species()._short_repr(),
            self._extracellular()._region._short_repr(),
        )

    def __str__(self):
        return "%s[%s]" % (
            self._species()._short_repr(),
            self._extracellular()._region._short_repr(),
        )

    @property
    def concentration(self):
        """An iterable of the current concentrations."""
        return self.nodes.concentration

    @concentration.setter
    def concentration(self, value):
        """Sets all Node objects in the restriction to the specified concentration value.

        This is equivalent to s.nodes.concentration = value"""
        self.nodes.concentration = value

    @property
    def states3d(self):
        return self._extracellular().states

    def extent(self, axes):
        """valid options for axes: xyz, xy, xz, yz, x, y, z"""
        e = self._extracellular()
        if axes == "xyz":
            return [e._xlo, e._xhi, e._ylo, e._yhi, e._zlo, e._zhi]
        elif axes == "xy":
            return [e._xlo, e._xhi, e._ylo, e._yhi]
        elif axes == "xz":
            return [e._xlo, e._xhi, e._zlo, e._zhi]
        elif axes == "yz":
            return [e._ylo, e._yhi, e._zlo, e._zhi]
        elif axes == "x":
            return [e._xlo, e._xhi]
        elif axes == "y":
            return [e._ylo, e._yhi]
        elif axes == "z":
            return [e._zlo, e._zhi]
        else:
            raise RxDException(
                "unknown axes argument; valid options: xyz, xy, xz, yz, x, y and z"
            )

    def node_by_location(self, x, y, z):
        e = self._extracellular()._region
        if e._xlo <= x <= e._xhi and e._ylo <= y <= e._yhi and e._zlo <= z <= e._zhi:
            i = int((x - e._xlo) / e._dx[0])
            j = int((y - e._ylo) / e._dx[1])
            k = int((z - e._zlo) / e._dx[2])
            return self.node_by_ijk(i, j, k)
        raise RxDException(
            "Location (%1.2f %1.2f, %1.2f) is not in extracellular region %r",
            x,
            y,
            z,
            self._extracellular()._region,
        )

    def alpha_by_location(self, locs):
        """Return a single alpha value for a homogeneous volume fraction of a list of alpha values for an inhomogeneous volume fraction at grid locations given in a list (locs)."""
        e = self._extracellular()._region
        if numpy.isscalar(e.alpha):
            return e.alpha
        alphas = []
        for loc in locs:
            i = int((loc[0] - e._xlo) / e._dx[0])
            j = int((loc[1] - e._ylo) / e._dx[1])
            k = int((loc[2] - e._zlo) / e._dx[2])
            alphas.append(e.alpha[i, j, k])
        return numpy.array(alphas)

    def node_by_ijk(self, i, j, k):
        index = 0
        s = self._extracellular()
        for ecs in self._species()._extracellular_instances.values():
            if ecs == s:
                e = s._region
                index += (i * e._ny + j) * e._nz + k
            else:
                e = ecs._region
                index += e._nz * e._ny * e._nz
        return self._species()._extracellular_nodes[index]

    @property
    def nodes(self):
        """A NodeList of the Node objects containing concentration data for the given Species and extracellular region.

        The code

            node_list = ca[ecs].nodes

        is more efficient than the otherwise equivalent

            node_list = ca.nodes(ecs)

        because the former only creates the Node objects belonging to the restriction ca[cyt] whereas the second option
        constructs all Node objects belonging to the Species ca and then culls the list to only include those also
        belonging to the Region cyt.
        """
        initializer._do_init()
        return nodelist.NodeList(
            [
                nd
                for nd in self._species()._extracellular_nodes
                if nd.region == self._extracellular()._region
            ]
        )

    def _semi_compile(self, reg, instruction):
        # This will always be an ecs_instance
        # reg = self._extracellular()._region
        # ecs_instance = self._species()._extracellular_instances[reg]
        # return ecs_instance._semi_compile(reg)
        return self._extracellular()._semi_compile(reg, instruction)

    @property
    def d(self):
        return self._extracellular()._d

    @d.setter
    def d(self, value):
        self._extracellular().d = value

    def defined_on_region(self, r):
        return r == self._extracellular()


class SpeciesOnRegion(_SpeciesMathable):
    def __init__(self, species, region):
        """The restriction of a Species to a Region."""
        self._species = weakref.ref(species)
        self._region = weakref.ref(region)

    def __eq__(self, other):
        """test for equality.

        Two SpeciesOnRegion objects are equal if they refer to the same species
        on the same region and both the species and the region still exist.
        """
        return (
            (self._species() == other._species())
            and (self._region() == other._region())
            and (self._species() is not None)
            and (self._region() is not None)
        )

    def __hash__(self):
        # TODO: replace this to reduce collision risk; how much of a problme is that?
        return 1000 * (hash(self._species()) % 1000) + (hash(self._region()) % 1000)

    def __repr__(self):
        return "%r[%r]" % (self._species(), self._region())

    def _short_repr(self):
        return "%s[%s]" % (self._species()._short_repr(), self._region()._short_repr())

    def __str__(self):
        return "%s[%s]" % (self._species()._short_repr(), self._region()._short_repr())

    def indices(self, r=None, secs=None):
        """If no Region is specified or if r is the Region specified in the constructor,
        returns a list of the indices of state variables corresponding
        to the Species when restricted to the Region defined in the constructor.

        If r is a different Region, then returns the empty list.
        If secs is a set return all indices on the regions for those sections.
        """
        # if r is not None and r != self._region():
        #    raise RxDException('attempt to access indices on the wrong region')
        # TODO: add a mechanism to catch if not right region (but beware of reactions crossing regions)
        if self._species() is None or self._region() is None:
            return []
        else:
            return self._species().indices(self._region(), secs)

    def __getitem__(self, r):
        """Return a reference to those members of this species lying on the specific region @varregion.
        The resulting object is a SpeciesOnRegion.
        This is useful for defining reaction schemes for MultiCompartmentReaction.

        This is provided to allow use of SpeciesOnRegion where Species would normally go.

        If the regions match, self is returned; otherwise an empty SpeciesOnRegion.
        """
        if r == self._region():
            return self
        raise RxDException("no such region")

    @property
    def states(self):
        """A vector of all the states corresponding to this species"""
        # TODO This should be restricted to the Region
        return self._species().states

    @property
    def states3D(self):
        """A vector of all the 3D states corresponding to this species"""
        # TODO This should be restricted to the Region
        return self._species()._intracellular_instances[self._region()].states

    @property
    def nodes(self):
        """A NodeList of the Node objects containing concentration data for the given Species and Region.

        The code

            node_list = ca[cyt].nodes

        is more efficient than the otherwise equivalent

            node_list = ca.nodes(cyt)

        because the former only creates the Node objects belonging to the restriction ca[cyt] whereas the second option
        constructs all Node objects belonging to the Species ca and then culls the list to only include those also
        belonging to the Region cyt.
        """
        from . import rxd

        _nrn_shape_update()
        if initializer.is_initialized():
            rxd._update_node_data()
        else:
            initializer._do_init()
        my_nodes = [
            s.nodes for s in self._species()._secs if s._region == self._region()
        ]
        try:
            my_nodes.append(self.instance3d._nodes)
        except:
            pass
        return nodelist.NodeList(itertools.chain.from_iterable(my_nodes))

    @property
    def concentration(self):
        """An iterable of the current concentrations."""
        return self.nodes.concentration

    @concentration.setter
    def concentration(self, value):
        """Sets all Node objects in the restriction to the specified concentration value.

        This is equivalent to s.nodes.concentration = value"""
        self.nodes.concentration = value

    def _semi_compile(self, reg, instruction):
        # Never an _ExtracellularSpecies since we are in SpeciesOnRegion
        reg = self._region()
        if instruction == "do_3d":
            ics_instance = self._species()._intracellular_instances[reg]
            return ics_instance._semi_compile(reg, instruction)
        elif any(reg._secs1d):
            return "species[%d][%d]" % (self._id, self._region()._id)
        else:
            raise RxDException("There are no 1D sections defined on {}".format(reg))

    def defined_on_region(self, r):
        return r == self._region()

    @property
    def instance3d(self):
        sp = self._species()
        r = self._region()
        if r in sp._intracellular_instances:
            return sp._intracellular_instances[r]
        else:
            raise RxDException("There are no 3D species defined on {}".format(r))

    @property
    def _id(self):
        return self._species()._id


# 3d matrix stuff
def _setup_matrices_process_neighbors(
    pt1, pt2, indices, euler_matrix, index, diffs, vol, areal, arear, dx
):
    # TODO: validate this before release! is ignoring reflective boundaries the right thing to do?
    #       (make sure that any changes here also work with boundaries that aren't really reflective, but have a 1d section attached)
    d = diffs[index]
    if pt1 in indices:
        ileft = indices[pt1]
        dleft = (d + diffs[ileft]) * 0.5
        left = dleft * areal / (vol * dx)
        euler_matrix[index, ileft] += left
        euler_matrix[index, index] -= left
    if pt2 in indices:
        iright = indices[pt2]
        dright = (d + diffs[iright]) * 0.5
        right = dright * arear / (vol * dx)
        euler_matrix[index, iright] += right
        euler_matrix[index, index] -= right


def _xyz(seg):
    """Return the (x, y, z) coordinate of the center of the segment."""
    # TODO: this is very inefficient, especially since we're calling this for each segment not for each section; fix
    sec = seg.sec
    n3d = sec.n3d()
    x3d = [sec.x3d(i) for i in range(n3d)]
    y3d = [sec.y3d(i) for i in range(n3d)]
    z3d = [sec.z3d(i) for i in range(n3d)]
    arc3d = [sec.arc3d(i) for i in range(n3d)]
    return (
        numpy.interp([seg.x * sec.L], arc3d, x3d)[0],
        numpy.interp([seg.x * sec.L], arc3d, y3d)[0],
        numpy.interp([seg.x * sec.L], arc3d, z3d)[0],
    )


class _IntracellularSpecies(_SpeciesMathable):
    def __init__(
        self,
        region=None,
        d=0,
        name=None,
        charge=0,
        initial=0,
        nodes=0,
        is_diffusable=True,
        atolscale=1.0,
    ):
        """
        region = Intracellular object
        name = string of the name of the NEURON species (e.g. ca)
        d = diffusion constant
        charge = charge of the molecule (used for converting currents to concentration changes)
        initial = initial concentration
        NOTE: For now, only define this AFTER the morphology is instantiated. In the future, this requirement can be relaxed.
        TODO: remove this limitation
        """
        _intracellular_diffusion_objects[self] = None
        # ensure 3D points exist
        h.define_shape()
        self._region = region
        self._species = name
        self._name = name
        self._charge = charge
        self._dx = region.dx
        self._atolscale = atolscale
        self._nodes_length = len(self._region._xs)
        self._states = h.Vector(self._nodes_length)
        self.neighbors = self.create_neighbors_array(nodes, self._nodes_length).reshape(
            3 * self._nodes_length
        )
        self._initial = initial
        self.states = self._states.as_numpy()
        self._nodes = nodes

        dc, dgrid = self._parse_diffusion(d)
        if dc is not None:
            self._d = dc
        else:
            self._dgrid = dgrid
        self._x_line_defs = self.line_defs(self._nodes, "x", self._nodes_length)
        self._y_line_defs = self.line_defs(self._nodes, "y", self._nodes_length)
        self._z_line_defs = self.line_defs(self._nodes, "z", self._nodes_length)

        self._ordered_x_nodes = self.ordered_nodes(
            self._x_line_defs, "x", self.neighbors
        )
        self._ordered_y_nodes = self.ordered_nodes(
            self._y_line_defs, "y", self.neighbors
        )
        self._ordered_z_nodes = self.ordered_nodes(
            self._z_line_defs, "z", self.neighbors
        )

        self._alphas = self.create_alphas()

        if hasattr(self, "_dgrid"):
            self._grid_id = ICS_insert_inhom(
                0,
                self._states._ref_x[0],
                len(self.states),
                self.neighbors,
                self._x_line_defs,
                len(self._x_line_defs),
                self._y_line_defs,
                len(self._y_line_defs),
                self._z_line_defs,
                len(self._z_line_defs),
                self._dgrid,
                self._dx,
                is_diffusable,
                self._atolscale,
                self._alphas,
            )
        else:
            self._grid_id = ICS_insert(
                0,
                self._states._ref_x[0],
                len(self.states),
                self.neighbors,
                self._x_line_defs,
                len(self._x_line_defs),
                self._y_line_defs,
                len(self._y_line_defs),
                self._z_line_defs,
                len(self._z_line_defs),
                self._d,
                self._dx,
                is_diffusable,
                self._atolscale,
                self._alphas,
            )

        self._update_pointers()
        _defined_species_by_gid.append(self)

    def _isalive(self):
        if self not in _intracellular_diffusion_objects:
            raise RxDException("The intracellular species has been deleted.")

    def _delete(self):
        global _intracellular_diffusion_objects, _defined_species_by_gid
        if self in _defined_species_by_gid:
            _defined_species_by_gid.remove(self)
        if _intracellular_diffusion_objects:
            if self in _intracellular_diffusion_objects:
                del _intracellular_diffusion_objects[self]
                # remove the grid id
                for sp in _intracellular_diffusion_objects:
                    if sp._grid_id > self._grid_id:
                        sp._grid_id -= 1
                _delete_by_id(self._grid_id)
                # remove any node.include_flux for the extracellular species.
                from . import node

                newflux = {
                    "index": [],
                    "type": [],
                    "source": [],
                    "scale": [],
                    "region": [],
                }
                for idx, t, src, sc, rptr in zip(
                    node._node_fluxes["index"],
                    node._node_fluxes["type"],
                    node._node_fluxes["source"],
                    node._node_fluxes["scale"],
                    node._node_fluxes["region"],
                ):
                    if t != self._grid_id:
                        if t > self._grid_id:
                            newflux["type"].append(t)
                        else:
                            newflux["type"].append(t - 1)
                        newflux["index"].append(idx)
                        newflux["type"].append(t)
                        newflux["source"].append(src)
                        newflux["scale"].append(sc)
                        newflux["region"].append(region)
                node._has_node_fluxes = newflux != node._node_fluxes
                node._node_fluxes = newflux

                # if this is the only intracellular species using
                # node._point_indices, then remove the corresponding region.
                for sp in _intracellular_diffusion_objects:
                    if sp._region == self._region:
                        break
                else:
                    if self._region in node._point_indices:
                        del node._point_indices[self._region]

                nrn_dll_sym("structure_change_cnt", ctypes.c_int).value += 1

    # Line Definitions for each direction
    def line_defs(self, nodes, direction, nodes_length):
        self._isalive()
        line_defs = []

        indices = set(range(0, nodes_length))

        pos_neg_from_dir = {"x": (0, 1), "y": (2, 3), "z": (4, 5)}
        pos, neg = pos_neg_from_dir[direction]
        while indices:
            line_start = indices.pop()
            line_length = 1
            for my_dir in [pos, neg]:
                node = nodes[line_start]
                while (node.neighbors[my_dir] is not None) and indices:
                    to_remove = node._neighbors[my_dir]
                    node = nodes[to_remove]
                    indices.remove(to_remove)
                    line_length += 1

            line_start = node._index
            line_defs.append([line_start, line_length])

        # sort list for parallelization
        line_defs.sort(key=lambda x: x[1], reverse=True)
        line_defs = numpy.asarray(line_defs, dtype=int)
        line_defs = line_defs.reshape(2 * len(line_defs))
        return line_defs

    def ordered_nodes(self, p_line_defs, direction, neighbors):
        self._isalive()
        offset_from_dir = {"x": 0, "y": 1, "z": 2}
        offset = offset_from_dir[direction]
        ordered_nodes = []
        for i in range(0, len(p_line_defs) - 1, 2):
            current_node = p_line_defs[i]
            line_length = p_line_defs[i + 1]
            ordered_nodes.append(current_node)
            for j in range(1, line_length):
                current_node = neighbors[current_node * 3 + offset]
                ordered_nodes.append(current_node)
        return numpy.asarray(ordered_nodes)

    def create_neighbors_array(self, nodes, nodes_length):
        self._isalive()
        my_array = numpy.zeros((nodes_length, 3), dtype=int)
        for n in nodes:
            for i, ele in enumerate(n.neighbors[::2]):
                my_array[n._index, i] = ele if ele is not None else -1
        return my_array

    def create_alphas(self):
        self._isalive()
        alphas = [vol / self._dx**3 for vol in self._region._vol]
        return numpy.asarray(alphas, dtype=float)

    def _import_concentration(self):
        self._isalive()
        ion = "_ref_" + self._name + self._region._nrn_region
        for seg, nodes in zip(
            r._segs3d(self._region._nodes_by_seg.keys()),
            self._region._nodes_by_seg.values(),
        ):
            segptr = getattr(seg, ion)
            value = segptr[0]
            for node in nodes:
                self._nodes[node].concentration = value

    def _finitialize(self):
        # Updated - now it will initialize using Node3D
        # TODO: support more complicated initializations than just constants
        self._isalive()
        if self._initial is None:
            if self._region._nrn_region:
                self._import_concentration()
            else:
                self.states[:] = 0

    def _update_pointers(self):
        self._isalive()
        if options.concentration_nodes_3d == "surface":
            self._seg_to_surface_nodes = self._region._surface_nodes_by_seg
        elif options.concentration_nodes_3d == "all":
            self._seg_to_surface_nodes = self._region._nodes_by_seg
        else:
            raise RxDException(
                "options.concentration_nodes_3d must be 'surface' or 'all'"
            )
        grid_list_start = 0
        if self._nodes and self._name:
            nrn_region = self._region._nrn_region
            if nrn_region:
                ion_conc = "_ref_" + self._name + nrn_region
                segs = self._region._segs3d()
                self._concentration_neuron_pointers = [
                    getattr(segs[segidx], ion_conc)
                    for segidx in self._seg_to_surface_nodes.keys()
                ]
                self._surface_nodes_per_seg = [
                    nodes for nodes in self._seg_to_surface_nodes.values()
                ]
                self._surface_nodes_per_seg_start_indices = [
                    len(nodes) for nodes in self._surface_nodes_per_seg
                ]

                # cast to numpy arrays to avoid calling PyList_GET_ITEM on every element
                self._surface_nodes_per_seg = list(
                    itertools.chain.from_iterable(self._surface_nodes_per_seg)
                )
                self._surface_nodes_per_seg = numpy.asarray(
                    self._surface_nodes_per_seg, dtype=numpy.int64
                )

                self._surface_nodes_per_seg_start_indices = numpy.cumsum(
                    [0] + self._surface_nodes_per_seg_start_indices, dtype=numpy.int64
                )

                _ics_set_grid_concentrations(
                    grid_list_start,
                    self._grid_id,
                    self._surface_nodes_per_seg,
                    self._surface_nodes_per_seg_start_indices,
                    self._concentration_neuron_pointers,
                )

                if self._name is not None and self._charge != 0:
                    nrn_region_sign = {"i": -1, "o": 1}
                    sign = nrn_region_sign[nrn_region]
                    ion_curr = "_ref_" + nrn_region + self._name
                    tenthousand_over_charge_faraday = 10000.0 / (
                        self._charge * h.FARADAY
                    )
                    my_dx = self._dx
                    # TODO: this better
                    if not isinstance(my_dx, tuple):
                        my_dx = (my_dx, my_dx, my_dx)
                    self._current_neuron_pointers = [
                        getattr(seg, ion_curr)
                        for seg in self._region._segs3d(
                            self._seg_to_surface_nodes.keys()
                        )
                    ]
                    # These are in the same order as self._surface_nodes_per_seg so self._surface_nodes_per_seg_start_indices will work for this list as well
                    geom_area = [
                        sum(self._region.geometry.surface_areas1d(sec))
                        for sec in self._region._secs3d
                    ]
                    node_area = [node.surface_area for node in self._nodes]
                    node_vol = [node.volume for node in self._nodes]
                    scale = geom_area / sum(node_area)
                    scale_factors = [
                        sign * area * scale * tenthousand_over_charge_faraday / vol
                        for area, vol in zip(node_area, node_vol)
                    ]
                    self._scale_factors = numpy.asarray(scale_factors, dtype=float)
                    _ics_set_grid_currents(
                        grid_list_start,
                        self._grid_id,
                        self._current_neuron_pointers,
                        self._scale_factors,
                    )

    def _semi_compile(self, region, instruction):
        self._isalive()
        if instruction == "do_3d":
            if self._species:
                sp = _defined_species[self._species][self._region]()
            else:
                for s in _all_species:
                    if self in s()._intracellular_instances.values():
                        sp = s()
                        break
            if isinstance(sp, Parameter):
                return "params_3d[%d]" % (self._grid_id)
            else:
                return "species_3d[%d]" % (self._grid_id)
        raise RxDException("_semi_compile received inconsistent state")

    def _register_cptrs(self):
        self._isalive()
        self._seg_order = []
        self._concentration_ptrs = []
        if self._nodes and self._species is not None:
            nrn_region = self._region._nrn_region
            if nrn_region is not None:
                ion = "_ref_" + self._species + nrn_region
                self._seg_order = list(self._region._nodes_by_seg.keys())
                for seg in self._region._segs3d(self._seg_order):
                    self._concentration_ptrs.append(getattr(seg, ion))

    def _import_concentration(self, init=True):
        self._isalive()
        nodes = self._nodes
        if not hasattr(self, "_seg_order"):
            self._register_cptrs()
        if nodes:
            # TODO: replace this with a pointer vec for speed
            #       not a huge priority since import happens rarely if at all
            conc_ptr = self._concentration_ptrs
            if self._region._nrn_region is not None:
                for segidx, ptr in zip(self._seg_order, conc_ptr):
                    value = ptr[0]
                    for node in self._region._nodes_by_seg[segidx]:
                        nodes[node].concentration = value

    def _transfer_to_legacy(self):
        self._isalive()
        nodes = self._nodes
        if self._nodes and self._region.nrn_region is not None:
            if self._region.nrn_region != "i":
                raise RxDException('only "i" nrn_region supported for 3D simulations')
                # TODO: remove this requirement
                #       the issue here is that the code below would need to keep track of which nodes are in which nrn_region
                #       that's not that big of a deal, but when this was coded, there were other things preventing this from working
        for segidx, ptr in zip(self._seg_order, self._concentration_ptrs):
            all_nodes_in_seg = list(self._region._nodes_by_seg[segidx])
            if all_nodes_in_seg:
                # TODO: if everything is 3D, then this should always have something, but for sections that aren't in 3D, won't have anything here
                # TODO: vectorize this, don't recompute denominator unless a structure change event happened
                ptr[0] = sum(
                    nodes[node].concentration * nodes[node].volume
                    for node in all_nodes_in_seg
                ) / sum(nodes[node].volume for node in all_nodes_in_seg)

    def _parse_diffusion(self, d):
        self._isalive()

        def get_signature(fun):
            import sys
            import inspect

            if sys.version_info.major > 2:
                sig = inspect.signature(fun)
                for param in sig.parameters.values():
                    if (
                        param.kind == param.VAR_POSITIONAL
                        or param.kind == param.VAR_KEYWORD
                    ):
                        raise RxDException(
                            "Intracellular diffusion coefficient function may not include *args or *kwargs"
                        )
                return len(sig.parameters)
            else:
                sig = inspect.getargspec(fun)
                if sig.varargs is not None or sig.keywords is not None:
                    raise RxDException(
                        "Intracellular diffusion coefficient function may not include *args or *kwargs"
                    )
                return len(sig.args)

        dc = None
        dgrid = None
        if callable(d) or (hasattr(d, "__len__") and callable(d[0])):
            dgrid = numpy.ndarray((3, self._nodes_length), dtype=float, order="C")
            if hasattr(d, "__len__"):
                if len(d) == 3:
                    for dr in range(3):
                        sig = get_signature(d[dr])
                        if sig == 3:
                            dgrid[dr, :] = [
                                d[dr](nd.x3d, nd.y3d, nd.z3d) for nd in self._nodes
                            ]
                        elif sig == 1:
                            dgrid[dr, :] = [d[dr](nd) for nd in self._nodes]
                else:
                    raise RxDException(
                        "Intracellular diffusion coefficient may be a scalar or a tuple of length 3 for anisotropic diffusion, it can also be a function for inhomogeneous diffusion (or tuple of 3 functions) with arguments x, y, z location or node with optional argument for direction"
                    )
            else:
                sig = get_signature(d)
                if sig == 4:
                    # x,y,z and direction
                    for dr in range(3):
                        dgrid[i, :] = [
                            d(nd.x3d, nd.y3d, nd.z3d, dr) for nd in self._nodes
                        ]

                elif sig == 3:
                    # x,y,z - isotropic
                    dg = [d(nd.x3d, nd.y3d, nd.z3d) for nd in self._nodes]
                    dgrid[0, :] = [d(nd.x3d, nd.y3d, nd.z3d) for nd in self._nodes]
                    dgrid[1, :] = dgrid[0, :]
                    dgrid[2, :] = dgrid[0, :]
                elif sig == 2:
                    # node and direction
                    for dr in range(3):
                        dgrid[dr, :] = [d(nd, dr) for nd in self._nodes]
                elif sig == 1:
                    # node - isotropic
                    dgrid[0, :] = [d(nd) for nd in self._nodes]
                    dgrid[1, :] = dgrid[0, :]
                    dgrid[2, :] = dgrid[0, :]
                else:
                    raise RxDException(
                        "Intracellular diffusion coefficient may be a scalar or a tuple of length 3 for anisotropic diffusion, it can also be a function for inhomogeneous diffusion (or tuple of 3 functions) with arguments x, y, z location or node with optional argument for direction"
                    )
        elif hasattr(d, "__len__"):
            if len(d) == 3:
                dc = numpy.array(d, dtype=float)
            elif len(d) == 1:
                dc = numpy.array(d[0] * numpy.ones(3), dtype=float)
            else:
                raise RxDException(
                    "Intracellular diffusion coefficient may be a scalar or a tuple of length 3 for anisotropic diffusion, it can also be a function for inhomogeneous diffusion (or tuple of 3 functions) with arguments x, y, z location or node with optional argument for direction"
                )
        else:
            dc = numpy.array(d * numpy.ones(3), dtype=float)
        return (dc, dgrid)

    @property
    def d(self):
        self._isalive()
        if hasattr(self, "_dgrid"):
            return self._dgrid
        return self._d

    @d.setter
    def d(self, value):
        self._isalive()
        dc, dgrid = self._parse_diffusion(value)
        if dc is not None:
            if hasattr(self, "_dgrid"):
                delattr(self, "_dgrid")
            self._d = dc
            set_diffusion(0, self._grid_id, dc, 1)
        else:
            self._dgrid = dgrid
            set_diffusion(0, self._grid_id, self._dgrid, self._nodes_length)

    # TODO Can we do this better?
    def _index_from_point(self, point):
        self._isalive()
        for i, n in enumerate(self._nodes):
            if point == (n._i, n._j, n._k):
                return i

    # TODO Can I do this better?
    def _mc3d_indices_start(self, r):
        self._isalive()
        indices = []
        for node_idx in r._nodes_by_seg.values():
            first_index = node_idx[0]
            point = r._points[first_index]
            indices.append(self._index_from_point(point))
        return min(indices)


class _ExtracellularSpecies(_SpeciesMathable):
    def __init__(
        self,
        region,
        d=0,
        name=None,
        charge=0,
        initial=0,
        atolscale=1.0,
        boundary_conditions=None,
        species=None,
    ):
        """
        region = Extracellular object (TODO? or list of objects)
        name = string of the name of the NEURON species (e.g. ca)
        d = diffusion constant
        charge = charge of the molecule (used for converting currents to concentration changes)
        initial = initial concentration
        alpha = volume fraction - either a single value for the whole region, a Vector giving a value for each voxel  or a anonymous function with argument of rxd.ExtracellularNode
        tortuosity = increase in path length due to obstacles, effective diffusion coefficient d/tortuosity^2. - either a single value for the whole region or a Vector giving a value for each voxel, or a anonymous function with argument of rxd.ExtracellularNode
        atolscale = scale factor for absolute tolerance in variable step integrations
        boundary_conditions = None by default corresponding to Neumann (zero flux) boundaries or a concentration for Dirichlet boundaries
        NOTE: For now, only define this AFTER the morphology is instantiated. In the future, this requirement can be relaxed.
        TODO: remove this limitation


        """
        _extracellular_diffusion_objects[self] = None

        # ensure 3D points exist
        h.define_shape()

        self._region = region
        self._species = name
        self._charge = charge
        self._xlo, self._ylo, self._zlo = region._xlo, region._ylo, region._zlo
        self._dx = region._dx
        self._d = d
        self._nx = region._nx
        self._ny = region._ny
        self._nz = region._nz
        self._xhi, self._yhi, self._zhi = region._xhi, region._yhi, region._zhi
        self._atolscale = atolscale
        self._states = h.Vector(self._nx * self._ny * self._nz)

        self.states = self._states.as_numpy().reshape(self._nx, self._ny, self._nz)
        self._initial = initial
        self._boundary_conditions = boundary_conditions
        self._diffusion_characteristic = (
            isinstance(region.alpha, Species) and region._alpha == species
        ) or (
            hasattr(region, "_permeability")
            and isinstance(region._permeability, Species)
            and region._permeability == species
        )
        if self._diffusion_characteristic:
            self.alpha = 1.0
            self._alpha = 1.0
        elif numpy.isscalar(region._volume_fraction_vector):
            self.alpha = self._alpha = region._volume_fraction_vector
        else:
            self.alpha = region.alpha
            self._alpha = region._volume_fraction_vector._ref_x[0]

        if self._diffusion_characteristic:
            self.tortuosity = 1.0
            self._permability = 1.0
        else:
            self.tortuosity = region.tortuosity
            if numpy.isscalar(region._permeability_vector):
                self._permability = region._permeability_vector
            else:
                self._permability = region._permeability_vector._ref_x[0]

        if boundary_conditions is None:
            bc_type = 0
            bc_value = 0.0
        else:
            bc_type = 1
            bc_value = boundary_conditions

        # TODO: if allowing different diffusion rates in different directions, verify that they go to the right ones
        if not hasattr(self._d, "__len__"):
            self._grid_id = ECS_insert(
                0,
                self._states._ref_x[0],
                self._nx,
                self._ny,
                self._nz,
                self._d,
                self._d,
                self._d,
                self._dx[0],
                self._dx[1],
                self._dx[2],
                self._alpha,
                self._permability,
                bc_type,
                bc_value,
                atolscale,
            )
        elif len(self._d) == 3:
            self._grid_id = ECS_insert(
                0,
                self._states._ref_x[0],
                self._nx,
                self._ny,
                self._nz,
                self._d[0],
                self._d[1],
                self._d[2],
                self._dx[0],
                self._dx[1],
                self._dx[2],
                self._alpha,
                self._permability,
                bc_type,
                bc_value,
                atolscale,
            )
        else:
            raise RxDException(
                "Diffusion coefficient %s for %s is invalid. A single value D or a tuple of 3 values (Dx,Dy,Dz) is required for the diffusion coefficient."
                % (repr(d), name)
            )

        self._name = name

        # set up the ion mechanism and enable active Nernst potential calculations
        self._ion_register()

        # moved to _finitialize -- in case the species is created before all the sections are
        # self._update_pointers()

        _defined_species_by_gid.append(self)

    def _isalive(self):
        if self not in _extracellular_diffusion_objects:
            raise RxDException("The extracellular species has been deleted.")

    def _delete(self):
        global _extracellular_diffusion_objects, _defined_species_by_gid
        # TODO: remove this object from the list of grids, possibly by reinserting all the others
        # NOTE: be careful about doing the right thing at program's end; some globals may no longer exist
        if self in _defined_species_by_gid:
            _defined_species_by_gid.remove(self)
        if _extracellular_diffusion_objects:
            if self in _extracellular_diffusion_objects:
                del _extracellular_diffusion_objects[self]
                # remove the grid id
                for sp in _extracellular_diffusion_objects:
                    if sp._grid_id > self._grid_id:
                        sp._grid_id -= 1
                _delete_by_id(self._grid_id)
                # remove any node.include_flux for the extracellular species.
                from . import node

                newflux = {
                    "index": [],
                    "type": [],
                    "source": [],
                    "scale": [],
                    "region": [],
                }
                for idx, t, src, sc, rptr in zip(
                    node._node_fluxes["index"],
                    node._node_fluxes["type"],
                    node._node_fluxes["source"],
                    node._node_fluxes["scale"],
                    node._node_fluxes["region"],
                ):
                    if t != self._grid_id:
                        if t > self._grid_id:
                            newflux["type"].append(t)
                        else:
                            newflux["type"].append(t - 1)
                        newflux["index"].append(idx)
                        newflux["type"].append(t)
                        newflux["source"].append(src)
                        newflux["scale"].append(sc)
                        newflux["region"].append(region)
                node._has_node_fluxes = newflux != node._node_fluxes
                node._node_fluxes = newflux
                nrn_dll_sym("structure_change_cnt", ctypes.c_int).value += 1

    def _finitialize(self):
        self._isalive()
        # Updated - now it will initialize using NodeExtracellular
        if self._initial is None:
            if hasattr(h, "%so0_%s_ion" % (self._species, self._species)):
                self.states[:] = getattr(
                    h, "%so0_%s_ion" % (self._species, self._species)
                )
            else:
                self.states[:] = 0
        if self._boundary_conditions:
            for idx in [0, -1]:
                self.states[idx, :, :] = self._boundary_conditions
                self.states[:, idx, :] = self._boundary_conditions
                self.states[:, :, idx] = self._boundary_conditions
        # necessary if section where created after the _ExtracellularSpecies
        self._update_pointers()

    def _ion_register(self):
        """modified from neuron.rxd.species.Species._ion_register"""

        self._isalive()
        if self._species is not None:
            ion_type = h.ion_register(self._species, self._charge)
            if ion_type == -1:
                raise RxDException("Unable to register species: %s" % self._species)
            # insert the species if not already present
            ion_forms = [
                self._species + "i",
                self._species + "o",
                "i" + self._species,
                "e" + self._species,
            ]
            for s in h.allsec():
                try:
                    for i in ion_forms:
                        # this throws an exception if one of the ion forms is missing
                        temp = getattr(s, i)
                except:
                    s.insert(self._species + "_ion")
                # set to recalculate reversal potential automatically
                # the last 1 says to set based on global initial concentrations
                # e.g. nao0_na_ion, etc...
                # TODO: this is happening GLOBALLY even to sections that aren't inside the extracellular domain (FIX THIS)
                h.ion_style(self._species + "_ion", 3, 2, 1, 1, 1, sec=s)

    def _nodes_by_location(self, i, j, k):
        self._isalive()
        return (i * self._ny + j) * self._nz + k

    def index_from_xyz(self, x, y, z):
        """Given an (x, y, z) point, return the index into the _species vector containing it or None if not in the domain"""

        self._isalive()
        if (
            x < self._xlo
            or x > self._xhi
            or y < self._ylo
            or y > self._yhi
            or z < self._zlo
            or z > self._zhi
        ):
            return None
        # if we make it here, then we are inside the domain
        i = int((x - self._xlo) / self._dx[0])
        j = int((y - self._ylo) / self._dx[1])
        k = int((z - self._zlo) / self._dx[2])
        return self._nodes_by_location(i, j, k)

    def ijk_from_index(self, index):
        self._isalive()
        nynz = self._ny * self._nz
        i = int(index / nynz)
        jk = index - nynz * i
        j = int(jk / self._nz)
        k = jk % self._nz
        # sanity check
        assert index == self._nodes_by_location(i, j, k)
        return i, j, k

    def xyz_from_index(self, index):
        self._isalive()
        i, j, k = ijk_from_index(self, index)
        return (
            self._xlo + i * self._dx[0],
            self._ylo + j * self._dx[1],
            self._zlo + k * self._dx[2],
        )

    def _locate_segments(self):
        """Note: there may be Nones in the result if a section extends outside the extracellular domain

        Note: the current version keeps all the sections alive (i.e. they will never be garbage collected)
        TODO: fix this
        """

        self._isalive()
        result = {}
        for sec in h.allsec():
            result[sec] = [self.index_from_xyz(*_xyz(seg)) for seg in sec]
        return result

    def _update_pointers(self):
        # TODO: call this anytime the _grid_id changes and anytime the structure_change_count change
        self._isalive()
        if self._species is None:
            return
        self._seg_indices = self._locate_segments()
        from .geometry import _surface_areas1d

        grid_list = 0
        grid_indices = []
        neuron_pointers = []
        stateo = "_ref_" + self._species + "o"
        for sec, indices in self._seg_indices.items():
            for seg, i in zip(sec, indices):
                if i is not None:
                    grid_indices.append(i)
                    neuron_pointers.append(getattr(seg, stateo))
        _set_grid_concentrations(
            grid_list, self._grid_id, grid_indices, neuron_pointers
        )
        if self._charge == 0 or isinstance(
            _defined_species[self._species][self._region](), Parameter
        ):
            _set_grid_currents(grid_list, self._grid_id, [], [], [])
        else:
            tenthousand_over_charge_faraday = 10000.0 / (self._charge * h.FARADAY)
            scale_factor = tenthousand_over_charge_faraday / (numpy.prod(self._dx))
            ispecies = "_ref_i" + self._species
            neuron_pointers = []
            scale_factors = []
            for sec, indices in self._seg_indices.items():
                for seg, surface_area, i in zip(sec, _surface_areas1d(sec), indices):
                    if i is not None:
                        neuron_pointers.append(getattr(seg, ispecies))
                        scale_factors.append(float(scale_factor * surface_area))
            # TODO: MultiCompartment reactions ?
            _set_grid_currents(
                grid_list, self._grid_id, grid_indices, neuron_pointers, scale_factors
            )

    def _import_concentration(self):
        """set node values to those of the corresponding NEURON segments"""

        self._isalive()
        if self._species is None:
            return
        grid_list = 0
        grid_indices = []
        neuron_pointers = []
        stateo = self._species + "o"
        for sec, indices in self._seg_indices.items():
            for seg, i in zip(sec, indices):
                if i is not None:
                    self._states[i] = getattr(seg, stateo)

    def _semi_compile(self, reg, instruction):

        self._isalive()
        if self._species:
            sp = _defined_species[self._species][self._region]()
        else:
            for s in _all_species:
                if self in s()._extracellular_instances.values():
                    sp = s()
                    break
        if isinstance(sp, Parameter):
            return "params_3d[%d]" % (self._grid_id)
        else:
            return "species_3d[%d]" % (self._grid_id)

    @property
    def d(self):

        self._isalive()
        return self._d

    @d.setter
    def d(self, value):
        self._isalive()
        if self._d != value:
            self._d = value
            if hasattr(value, "__len__"):
                set_diffusion(0, self._grid_id, numpy.array(value, dtype=float), 1)
            else:
                set_diffusion(0, self._grid_id, numpy.repeat(float(value), 3), 1)


# TODO: make sure that we can make this work where things diffuse across the
#       boundary between two regions... for the tree solver, this is complicated
#       because need the sections in a sorted order
#       ... but this is also weird syntactically, because how should it know
#       that two regions (e.g. apical and other_dendrites) are connected?


class Species(_SpeciesMathable):
    def __init__(
        self,
        regions=None,
        d=0,
        name=None,
        charge=0,
        initial=None,
        atolscale=1,
        ecs_boundary_conditions=None,
        represents=None,
    ):
        """s = rxd.Species(regions, d = 0, name = None, charge = 0, initial = None, atolscale=1)

        Declare a species.

        Parameters:

        regions -- a Region or list of Region objects containing the species

        d -- the diffusion constant of the species (optional; default is 0, i.e. non-diffusing)

        name -- the name of the Species; used for syncing with NMODL and HOC (optional; default is none)

        charge -- the charge of the Species (optional; default is 0)

        initial -- the initial concentration or None (if None, then imports from HOC if the species is defined at finitialize, else 0)

        atolscale -- scale factor for absolute tolerance in variable step integrations

        ecs_boundary_conditions  -- if Extracellular rxd is used ecs_boundary_conditions=None for zero flux boundaries or if ecs_boundary_conditions=the concentration at the boundary.

        represents -- optionally provide CURIE (Compact URI) to annotate what the species represents e.g. CHEBI:29101 for sodium(1+)

        Note:

        charge must match the charges specified in NMODL files for the same ion, if any.

        You probably want to adjust atolscale for species present at low concentrations (e.g. calcium).
        """

        import neuron
        import ctypes

        from . import rxd, region

        # if there is a species, then rxd is being used, so we should register
        # this function may be safely called many times
        rxd._do_nbs_register()

        # TODO: check if "name" already defined elsewhere (if non-None)
        #       if so, make sure other fields consistent, expand regions as appropriate

        self._allow_setting = True
        self.regions = regions
        # Check for anisotropic/inhomogeneous 3D diffusion with 1D sections
        if hasattr(d, "__len__") or callable(d):
            if (
                hasattr(d, "__len__")
                and all([di == d[0] for di in d])
                and not callable(d[0])
            ):
                self._d = d[0]
            else:
                from .rxd import _domain_lookup

                reglist = regions if hasattr(regions, "__len__") else [regions]
                dims = [
                    _domain_lookup(sec)
                    for reg in reglist
                    if not isinstance(reg, region.Extracellular)
                    for sec in reg.secs
                ]
                if not all([dim == dims[0] for dim in dims]):
                    raise RxDException(
                        "Hybrid 1D/3D diffusion does not currently support anisotropy or inhomogeneous grids. For separate 1D and 3D diffusion please create separate regions and species for 3D and 1D sections"
                    )
                else:
                    self._d = d
        else:
            self._d = d
        self.name = name
        self.charge = charge
        self.initial = initial
        self._atolscale = atolscale
        self._ecs_boundary_conditions = ecs_boundary_conditions
        self._species_on_region = weakref.WeakKeyDictionary()
        if represents and not _ontology_id.match(represents):
            raise RxDException("The represents=%s is not valid CURIE" % represents)
        else:
            self.represents = represents
        with initializer._init_lock:
            _all_species.append(weakref.ref(self))

            # declare an update to the structure of the model (the number of differential equations has changed)
            nrn_dll_sym("structure_change_cnt", ctypes.c_int).value += 1

            self._ion_register()

            # initialize self if the rest of rxd is already initialized
            if initializer.is_initialized():
                self._do_init()
                rxd._update_node_data(True, True)

    def _do_init(self):
        self._do_init1()
        self._do_init2()
        self._do_init3()
        self._do_init4()
        self._do_init5()
        self._do_init6()

    def _do_init1(self):
        from . import rxd

        # TODO: if a list of sections is passed in, make that one region
        # _species_count is used to create a unique _real_name for the species
        global _species_count

        regions = self.regions
        self._real_name = self._name
        initial = self.initial
        charge = self._charge

        # invalidate any old initialization of external solvers
        rxd._external_solver_initialized = False

        # TODO: check about the 0<x<1 problem alluded to in the documentation
        h.define_shape()
        name = self.name
        if name is not None:
            if not isinstance(name, str):
                raise RxDException("Species name must be a string")
            if name in _defined_species:
                spsecs_i = []
                spsecs_o = []
                for r in regions:
                    if r in _defined_species[name]:
                        raise RxDException(
                            'Species "%s" previously defined on region: %r' % (name, r)
                        )
                    if hasattr(r, "_secs"):
                        if r.nrn_region == "i":
                            spsecs_i += r._secs
                        elif r.nrn_region == "o":
                            spsecs_o += r._secs
                spsecs_i = set(spsecs_i)
                spsecs_o = set(spsecs_o)
                for r in _defined_species[name]:
                    if hasattr(r, "_secs") and r.nrn_region:
                        if (
                            r.nrn_region == "i"
                            and any(spsecs_i.intersection(r._secs))
                            or (
                                r.nrn_region == "o"
                                and any(spsecs_o.intersection(r._secs))
                            )
                        ):
                            raise RxDException(
                                'Species "%s" previously defined on a region %r that overlaps with regions: %r'
                                % (name, r, self._regions)
                            )
        else:
            name = _species_count
        self._id = _species_count
        _species_count += 1
        if name not in _defined_species:
            _defined_species[name] = weakref.WeakKeyDictionary()
        self._species = weakref.ref(self)
        for r in regions:
            _defined_species[name][r] = self._species
        _all_defined_species.append(self._species)

        if regions is None:
            raise RxDException("Must specify region where species is present")
        if hasattr(regions, "__len__"):
            regions = list(regions)
        else:
            regions = list([regions])
        # TODO: unite handling of _regions and _extracellular_regions
        self._regions = [r for r in regions if not isinstance(r, region.Extracellular)]
        self._extracellular_regions = [
            r for r in regions if isinstance(r, region.Extracellular)
        ]
        if not all(isinstance(r, region.Region) for r in self._regions):
            raise RxDException(
                "regions list must consist of Region and Extracellular objects only"
            )

        # at this point self._name is None if unnamed or a string == name if
        # named
        self._ion_register()

        # TODO: remove this line when certain no longer need it (commented out 2013-04-17)
        # self._real_secs = region._sort_secs(sum([r.secs for r in regions], []))

    def _do_init2(self):
        # 1D section objects; NOT all sections this species lives on
        # TODO: this may be a problem... debug thoroughly
        global _has_1d
        d = self._d

        self._secs = [
            Section1D(self, sec, d, r) for r in self._regions for sec in r._secs1d
        ]
        if self._secs:
            self._offset = self._secs[0]._offset
            _has_1d = True
        else:
            self._offset = 0
        self._has_adjusted_offsets = False
        self._assign_parents()

    def _do_init3(self):
        global _has_3d
        # 3D sections

        # NOTE: if no 3D nodes, then _3doffset is not meaningful
        self._3doffset = 0
        self._3doffset_by_region = {}
        selfref = weakref.ref(self)
        self_has_3d = False
        self._intracellular_nodes = {}
        if self._regions:
            for r in self._regions:
                self._intracellular_nodes[r] = []
                if any(r._secs3d):
                    xs, ys, zs, segsidx = r._xs, r._ys, r._zs, r._segsidx
                    segs = [seg for sec in r._secs3d for seg in sec]
                    self._intracellular_nodes[r] += [
                        node.Node3D(i, x, y, z, r, self._d, segs[idx], selfref)
                        for i, x, y, z, idx in zip(range(len(xs)), xs, ys, zs, segsidx)
                    ]
                    # the region is now responsible for computing the correct volumes and surface areas
                    # this is done so that multiple species can use the same region without recomputing it
                    self_has_3d = True
                    _has_3d = True
        is_diffusable = (
            False if isinstance(self, Parameter) or isinstance(self, State) else True
        )
        self._intracellular_instances = {
            r: _IntracellularSpecies(
                r,
                d=self._d,
                charge=self.charge,
                initial=self.initial,
                nodes=self._intracellular_nodes[r],
                name=self._name,
                is_diffusable=is_diffusable,
                atolscale=self._atolscale,
            )
            for r in self._regions
            if any(r._secs3d)
        }

    def _do_init4(self):
        extracellular_nodes = []
        self._extracellular_instances = {
            r: _ExtracellularSpecies(
                r,
                d=self._d,
                name=self.name,
                charge=self.charge,
                initial=self.initial,
                atolscale=self._atolscale,
                boundary_conditions=self._ecs_boundary_conditions,
                species=self,
            )
            for r in self._extracellular_regions
        }
        sp_ref = weakref.ref(self)
        index = 0
        for r in self._extracellular_regions:
            r_ref = weakref.ref(r)
            for i in range(r._nx):
                for j in range(r._ny):
                    extracellular_nodes += [
                        node.NodeExtracellular(idx, i, j, k, sp_ref, r_ref)
                        for k, idx in enumerate(range(index, index + r._nz))
                    ]
                    index += r._nz
            # extracellular_nodes += [node.NodeExtracellular(idx, i, j, k, sp_ref, r_ref) for i, idxx in  enumerate(range(index, index + r._nx * r._ny * r._nz, r._ny * r._nz)) for j, idxy in enumerate(range(idxx, idxx + r._nz * r._ny, r._nz)) for k,idx in enumerate(range(idxy, idxy + r._nz))]
            # index = len(extracellular_nodes)
        self._extracellular_nodes = extracellular_nodes

    def _do_init5(self):
        # final initialization
        for sec in self._secs:
            # NOTE: can only init_diffusion_rates after the roots (parents)
            #       have been assigned
            sec._init_diffusion_rates()
        self._update_node_data()  # is this really nessesary
        self._update_region_indices()

    def _do_init6(self):
        self._register_cptrs()

    def __del__(self):
        global _all_species, _defined_species, _all_defined_species
        if hasattr(self, "deleted"):
            return
        self.deleted = True
        if hasattr(self, "_id"):
            remove_species_atolscale(self._id)

        name = self.name if self.name else (self._id if hasattr(self, "_id") else None)
        if name is not None and name in _defined_species:
            for r in self.regions:
                if r in _defined_species[name]:
                    del _defined_species[name][r]
            if not any(_defined_species[name]):
                del _defined_species[name]
        _all_defined_species = list(
            filter(lambda x: x() is not None and x() is not self, _all_defined_species)
        )
        _all_species = list(
            filter(lambda x: x() is not None and x() is not self, _all_species)
        )
        if hasattr(self, "_extracellular_instances"):
            for sp in self._extracellular_instances.values():
                sp._delete()

        if hasattr(self, "_intracellular_instances"):
            for sp in self._intracellular_instances.values():
                sp._delete()

        # delete the secs
        if hasattr(self, "_secs") and self._secs:
            # remove the species root
            # TODO: will this work with post initialization morphology changes
            self._secs.sort(key=lambda s: s._offset, reverse=True)
            for sec in self._secs[:]:
                sec._delete()
            self._secs = []

        # set the remaining species offsets
        if initializer.is_initialized():
            for sr in _all_species:
                s = sr()
                if s is not None:
                    s._update_region_indices(True)

    def _ion_register(self):
        charge = self._charge
        if self._name is not None:
            ion_type = None
            # insert the species if not already present
            regions = (
                self._regions if hasattr(self._regions, "__len__") else [self._regions]
            )
            for r in regions:
                if r.nrn_region in ("i", "o"):
                    if ion_type is None:
                        ion_type = h.ion_register(self._name, charge)
                        if ion_type == -1:
                            raise RxDException(
                                "Unable to register species: %s" % self._name
                            )

                    for s in r.secs:
                        try:
                            ion_forms = [
                                self._name + "i",
                                self._name + "o",
                                "i" + self._name,
                                "e" + self._name,
                            ]
                            for i in ion_forms:
                                # this throws an exception if one of the ion forms is missing
                                temp = getattr(s, self._name + "i")
                        except:
                            s.insert(self._name + "_ion")
                        # set to recalculate reversal potential automatically
                        # the last 1 says to set based on global initial concentrations
                        # e.g. nai0_na_ion, etc...
                        h.ion_style(self._name + "_ion", 3, 2, 1, 1, 1, sec=s)

    @property
    def states(self):
        """A list of all the state values corresponding to this species"""
        all_states = node._get_states()
        return [all_states[i] for i in numpy.sort(self.indices())]

    @property
    def _state(self):
        """return a bytestring representing the Species state"""
        # format: version identifier (unsigned long long), size (unsigned long long), binary data
        import array

        version = 0
        data = array.array("d", self.nodes.concentration).tobytes()
        return array.array("Q", [version, len(data)]).tobytes() + data

    @_state.setter
    def _state(self, oldstate):
        """restore Species state"""
        import array

        metadata_array = array.array("Q")
        metadata_array.frombytes(oldstate[:16])
        version, length = metadata_array
        if version != 0:
            raise RxdException("Unsupported state data version")
        data = array.array("d")
        try:
            data.frombytes(oldstate[16:])
        except ValueError:
            # happens when not a multiple of 8 bytes
            raise RxDException("Invalid state data length") from None
        # at 8 bytes per data point, the total number of bytes should match the stored
        if len(data) * 8 != length or len(data) != len(self.nodes):
            raise RxDException("Invalid state data length")
        self.nodes.concentration = data

    def re_init(self):
        """Reinitialize the rxd concentration of this species to match the NEURON grid"""
        self._import_concentration(init=False)

    def __getitem__(self, r):
        """Return a reference to those members of this species lying on the specific region @varregion.
        The resulting object is a SpeciesOnRegion.
        This is useful for defining reaction schemes for MultiCompartmentReaction."""
        if isinstance(r, region.Region) and r in self._regions:
            if r not in self._species_on_region:
                self._species_on_region[r] = SpeciesOnRegion(self, r)
            return self._species_on_region[r]
        elif isinstance(r, region.Extracellular):
            if not hasattr(self, "_extracellular_instances"):
                initializer._do_init()
            if r in self._extracellular_instances:
                if r not in self._species_on_region:
                    self._species_on_region[r] = SpeciesOnExtracellular(
                        self, self._extracellular_instances[r]
                    )
                return self._species_on_region[r]
        raise RxDException("no such region")

    def _update_node_data(self):
        nsegs_changed = 0
        for sec in self._secs:
            nsegs_changed += sec._update_node_data()
        # remove deleted sections
        self._secs = [sec for sec in self._secs if sec and sec._nseg > 0]

        # check if 3D sections/nseg has changed
        for r in self._intracellular_instances:
            if r._secs3d_names:
                secs3d_names = {sec.hoc_internal_name(): sec.nseg for sec in r._secs3d}
                if secs3d_names:
                    if secs3d_names != r._secs3d_names:
                        # TODO: redo voxelization and interpolate
                        raise RxDException(
                            "Error: changing the 3D sections or their nseg after initialization is not yet supported."
                        )
                else:
                    # If there are no 3D sections we can continue without
                    # re-voxelization.
                    self._intracellular_nodes[r] = []

        return nsegs_changed

    def concentrations(self):
        if self._secs:
            raise RxDException("concentrations only supports 3d and that is deprecated")
        warnings.warn("concentrations is deprecated; do not use")
        r = self._regions[0]
        data = numpy.array(r._mesh.values, dtype=float)
        # things outside of the volume will be NaN
        data[:] = numpy.NAN
        max_concentration = -1
        for node in self.nodes:
            data[node._i, node._j, node._k] = node.concentration
        return data

    def _register_cptrs(self):
        # 1D stuff
        for sec in self._secs:
            sec._register_cptrs()
        idx = self._indices1d()
        species_atolscale(
            self._id, self._atolscale, len(idx), (ctypes.c_int * len(idx))(*idx)
        )
        # 3D stuff
        for ics in self._intracellular_instances.values():
            ics._register_cptrs()

    @property
    def charge(self):
        """Get or set the charge of the Species.

        .. note:: Setting was added in NEURON 7.4+ and is allowed only before the reaction-diffusion model is instantiated.
        """
        return self._charge

    @charge.setter
    def charge(self, value):
        if hasattr(self, "_allow_setting"):
            self._charge = value
        else:
            raise RxDException("Cannot set charge now; model already instantiated")

    @property
    def regions(self):
        """Get or set the regions where the Species is present

        .. note:: New in NEURON 7.4+. Setting is allowed only before the reaction-diffusion model is instantiated.

        .. note:: support for getting the regions is new in NEURON 7.5.
        """
        return list(self._regions) + list(self._extracellular_regions)

    @regions.setter
    def regions(self, regions):
        if hasattr(self, "_allow_setting"):
            if hasattr(regions, "__len__"):
                regions = list(regions)
            else:
                regions = list([regions])
            self._regions = [
                r for r in regions if not isinstance(r, region.Extracellular)
            ]
            self._extracellular_regions = [
                r for r in regions if isinstance(r, region.Extracellular)
            ]
        else:
            raise RxDException("Cannot set regions now; model already instantiated")

    def _update_region_indices(self, update_offsets=False):
        # TODO: should this include 3D nodes? currently 1D only. (What faces the user?)
        # update offset in case nseg has changed
        self._region_indices = {}
        if hasattr(self, "_secs") and self._secs:
            if update_offsets:
                self._offset = self._secs[0]._offset - self._num_roots
            for s in self._secs:
                if s._region not in self._region_indices:
                    self._region_indices[s._region] = []
                self._region_indices[s._region] += s.indices
        # a list of all indices
        self._region_indices[None] = list(
            itertools.chain.from_iterable(list(self._region_indices.values()))
        )

    def _indices3d(self, r=None):
        """return the indices of just the 3D nodes corresponding to this species in the given region"""
        # TODO: Remove this
        return []
        # TODO: this will need changed if 3D is to support more than one region
        if r is None or r == self._regions[0]:
            return list(range(self._3doffset, self._3doffset + len(self._nodes)))
        else:
            return []

    def _indices1d(self, r=None, secs=None):
        """return the indices of just the 1D nodes corresponding to this species in the given region"""
        if secs is not None:
            if r is not None:
                return [
                    idx
                    for sec in self._secs
                    for idx in sec.indices
                    if sec in secs and sec._region == r
                ]
            return [idx for sec in self._secs for idx in sec.indices if sec in secs]
        return self._region_indices.get(r, [])

    def indices(self, r=None, secs=None):
        """return the indices corresponding to this species in the given region

        if r is None, then returns all species indices
        If r is a list of regions return indices for only those sections that are on all the regions.
        If secs is a set return all indices on the regions for those sections."""
        # TODO: beware, may really want self._indices3d or self._indices1d
        initializer._do_init()
        if secs is not None:
            if type(secs) != set:
                secs = {secs}
            if r is None:
                regions = self._regions
            elif not hasattr(r, "__len__"):
                regions = [r]
            else:
                regions = r
            return list(
                itertools.chain.from_iterable(
                    [
                        s.indices
                        for s in self._secs
                        if s._sec in secs and s.region in regions
                    ]
                )
            )

            """for reg in regions:
                ind = self.indices(reg)
                offset = 0
                for sec in r.secs:
                    if sec not in secs:
                        del ind[offset:(offset+sec.nseg)]
                    offset += sec.nseg
                return ind"""
        else:
            if not hasattr(r, "__len__"):
                return self._indices1d(r) + self._indices3d(r)
            elif len(r) == 1:
                return self._indices1d(r[0]) + self._indices3d(r[0])
            else:  # Find the intersection
                interseg = set.intersection(
                    *[set(reg.secs) for reg in r if reg is not None]
                )
                return self.indices(r, interseg)

    def defined_on_region(self, r):
        return r in self._regions or r in self._extracellular_regions

    def _setup_diffusion_matrix(self, g):
        for s in self._secs:
            s._setup_diffusion_matrix(g)

    def _setup_c_matrix(self, c):
        # TODO: this will need to be changed for three dimensions, or stochastic
        for s in self._secs:
            c[s._offset : s._offset + s.nseg] = 1.0

    def _setup_currents(self, indices, scales, ptrs, cur_map):
        from . import rxd

        if self.name:
            if self.name + "i" not in cur_map:
                cur_map[self.name + "i"] = {}
            if self.name + "o" not in cur_map:
                cur_map[self.name + "o"] = {}
        # 1D part
        for s in self._secs:
            s._setup_currents(indices, scales, ptrs, cur_map)

    def _has_region_section(self, region, sec):
        return any((s._region == region and s._sec == sec) for s in self._secs)

    def _region_section(self, region, sec):
        """return the Section1D (or whatever) associated with the region and section"""
        for s in self._secs:
            if s._region == region and s._sec == sec:
                return s
        else:
            raise RxDException("requested section not in species")

    def _assign_parents(self):
        root_id = 0
        missing = {}
        self._root_children = []
        for sec in self._secs:
            root_id = sec._assign_parents(root_id, missing, self._root_children)
        self._num_roots = root_id
        # TODO: this probably doesn't do the right thing if the morphology
        #       is changed after creation
        # adjust offsets to account for roots
        if not self._has_adjusted_offsets:
            node._allocate(self._num_roots)
            for sec in self._secs:
                sec._offset += root_id
            # used for section1d.__del__
            if self._secs != []:
                self._secs[0]._num_roots = self._num_roots
            self._has_adjusted_offsets = True

    def _finitialize(self, skip_transfer=False):
        if self._extracellular_instances:
            for r in self._extracellular_instances.keys():
                self._extracellular_instances[r]._finitialize()
        if self._intracellular_instances:
            for r in self._intracellular_instances.keys():
                self._intracellular_instances[r]._finitialize()
        if self.initial is not None:
            if isinstance(self.initial, Callable):
                for node in self.nodes:
                    node.concentration = self.initial(node)
            else:
                for node in self.nodes:
                    node.concentration = self.initial
            if not skip_transfer:
                self._transfer_to_legacy()
        else:
            self._import_concentration()

    def _transfer_to_legacy(self):
        """Transfer concentrations to the standard NEURON grid"""
        if self._name is None:
            return

        # 1D part
        for sec in self._secs:
            sec._transfer_to_legacy()

        # now the 3D stuff
        for ics in self._intracellular_instances.values():
            ics._transfer_to_legacy()

        """
        nodes = self._nodes
        if nodes:
            non_none_regions = [r for r in self._regions if r._nrn_region is not None]
            if non_none_regions:
                if any(r._nrn_region != 'i' for r in non_none_regions):
                    raise RxDException('only "i" nrn_region supported for 3D simulations')
                    # TODO: remove this requirement
                    #       the issue here is that the code below would need to keep track of which nodes are in which nrn_region
                    #       that's not that big of a deal, but when this was coded, there were other things preventing this from working                    
                for seg, ptr in zip(self._seg_order, self._concentration_ptrs):
                    all_nodes_in_seg = list(itertools.chain.from_iterable(r._surface_nodes_by_seg[seg] for r in non_none_regions))
                    if all_nodes_in_seg:
                        # TODO: if everything is 3D, then this should always have something, but for sections that aren't in 3D, won't have anything here
                        # TODO: vectorize this, don't recompute denominator unless a structure change event happened
                        ptr[0] = sum(nodes[node].concentration * nodes[node].volume for node in all_nodes_in_seg) / sum(nodes[node].volume for node in all_nodes_in_seg)
        """

    def _import_concentration(self, init=True):
        """Read concentrations from the standard NEURON grid"""
        if self._name is None:
            return

        # start with the 1D stuff
        for sec in self._secs:
            sec._import_concentration(init)

        # now the 3D stuff
        for ics in self._intracellular_instances.values():
            ics._import_concentration()

        # now the ECS
        for ecs in self._extracellular_instances.values():
            ecs._import_concentration()

    @property
    def nodes(self):
        """A NodeList of all the nodes corresponding to the species.

        This can then be further restricted using the callable property of NodeList objects."""

        from . import rxd

        _nrn_shape_update()
        if initializer.is_initialized():
            rxd._update_node_data()
        else:
            initializer._do_init()

        self._all_intracellular_nodes = []
        if self._intracellular_nodes:
            for r in self._regions:
                if r in self._intracellular_nodes:
                    self._all_intracellular_nodes += self._intracellular_nodes[r]
        # The first part here is for the 1D -- which doesn't keep live node objects -- the second part is for 3D
        self._all_intracellular_nodes = [
            nd for nd in self._all_intracellular_nodes[:] if nd.sec
        ]
        return nodelist.NodeList(
            list(itertools.chain.from_iterable([s.nodes for s in self._secs]))
            + self._all_intracellular_nodes
            + self._extracellular_nodes
        )

    @property
    def name(self):
        """Get or set the name of the Species.

        This is used only for syncing with the non-reaction-diffusion parts of NEURON.

        .. note:: Setting only supported in NEURON 7.4+, and then only before the reaction-diffusion model is instantiated.
        """
        return self._name

    @name.setter
    def name(self, value):
        if hasattr(self, "_allow_setting"):
            self._name = value
        else:
            raise RxDException("Cannot set name now; model already instantiated")

    def __repr__(self):
        represents = ", represents=%s" % self.represents if self.represents else ""
        return "Species(regions=%r, d=%r, name=%r, charge=%r, initial=%r%s)" % (
            self._regions,
            self._d,
            self._name,
            self._charge,
            self.initial,
            represents,
        )

    def _short_repr(self):
        if self._name is not None:
            return str(self._name)
        else:
            return self.__repr__()

    def __str__(self):
        if self._name is None:
            return self.__repr__()
        return str(self._name)


def xyz_by_index(indices):
    """Return the 3D location of the nodes at the given indices"""
    if hasattr(indices, "count"):
        index = indices
    else:
        index = [indices]
    # TODO: make sure to include Node3D
    return [
        [nd.x3d, nd.y3d, nd.z3d]
        for sp in _get_all_species()
        for s in sp()._secs
        for nd in s.nodes
        if sp()
        if nd._index in index
    ]


class Parameter(Species):
    """
    s = rxd.Parameter(regions, name=None, charge=0, value=None, represents=None)

    Declare a parameter, it can be used in place of a rxd.Species, but unlike rxd.Species a parameter will not change.

    Parameters:
    regions -- a Region or list of Region objects containing the parameter

    name -- the name of the parameter; used for syncing with NMODL and HOC (optional; default is none)

    charge -- the charge of the Parameter (optional; default is 0)

    value -- the value or None (if None, then imports from HOC if the parameter is defined at finitialize, else 0)

    represents -- optionally provide CURIE (Compact URI) to annotate what the parameter represents e.g. CHEBI:29101 for sodium(1+)

    Note:
    charge must match the charges specified in NMODL files for the same ion, if any.
    """

    def __init__(self, *args, **kwargs):
        if "value" in kwargs:
            if (
                "initial" in kwargs
                and kwargs["initial"]
                and kwargs["initial"] != kwargs["value"]
            ):
                raise RxDException(
                    "Parameter cannot be assigned both a 'value=%g' and 'initial=%g'"
                    % (kwargs["value"], kwargs["initial"])
                )
            kwargs["initial"] = kwargs.pop("value")
        if "d" in kwargs and kwargs["d"] != 0:
            raise RxDException(
                "Parameters cannot diffuse, invalid keyword argument 'd=%g'"
                % kwargs["d"]
            )
        self._d = 0
        super(Parameter, self).__init__(*args, **kwargs)

    def _semi_compile(self, reg, instruction):
        from . import region

        # region is Extracellular
        if isinstance(reg, region.Extracellular):
            ecs_instance = self._extracellular_instances[reg]
            return ecs_instance._semi_compile(reg, instruction)
        # region is 3d intracellular
        if isinstance(reg, region.Region) and instruction == "do_3d":
            ics_instance = self._intracellular_instances[reg]
            return ics_instance._semi_compile(reg, instruction)
        if isinstance(reg, region.Region) and instruction == "do_1d":
            if reg in self._regions:
                return "params[%d][%d]" % (self._id, reg._id)
            elif len(self._regions) == 1:
                return "params[%d][%d]" % (self._id, self._regions[0]._id)
            else:
                raise RxDException(
                    "Parameter %r is not defined on region %r" % (self, reg)
                )
        raise RxDException("_semi_compile received inconsistent state")

    def __repr__(self):
        represents = ", represents=%s" % self.represents if self.represents else ""
        return "Parameter(regions=%r, name=%r, charge=%r, value=%r%s)" % (
            self._regions,
            self._name,
            self._charge,
            self.initial,
            represents,
        )

    @property
    def value(self):
        return self.initial

    @value.setter
    def value(self, v):
        self.initial = v

    @property
    def d(self):
        return 0

    @d.setter
    def d(self, value):
        if value != 0:
            raise RxDException("Parameters cannot diffuse.")

    def __getitem__(self, r):
        """Return a reference to those members of this parameter lying on the specific region @varregion.
        The resulting object is a ParameterOnRegion or ParameterOnExtracellular.
        This is useful for defining reaction schemes for MultiCompartmentReaction."""
        if isinstance(r, region.Region) and r in self._regions:
            if r not in self._species_on_region:
                self._species_on_region[r] = ParameterOnRegion(self, r)
            return self._species_on_region[r]
        elif isinstance(r, region.Extracellular):
            if not hasattr(self, "_extracellular_instances"):
                initializer._do_init()
            if r in self._extracellular_instances:
                if r not in self._species_on_region:
                    self._species_on_region[r] = ParameterOnExtracellular(
                        self, self._extracellular_instances[r]
                    )
                return self._species_on_region[r]
        raise RxDException("no such region")


class ParameterOnRegion(SpeciesOnRegion):
    def _semi_compile(self, reg, instruction):
        """reg = self._region()
        if reg._secs3d:
            ics_instance = self._species()._intracellular_instances[reg]
            return ics_instance._semi_compile(reg, instruction)
        elif reg._secs1d:
            return 'params[%d][%d]' % (self._id, self._region()._id)"""
        reg = self._region()
        if instruction == "do_3d":
            ics_instance = self._species()._intracellular_instances[reg]
            return ics_instance._semi_compile(reg, instruction)
        elif any(reg._secs1d):
            return "params[%d][%d]" % (self._id, self._region()._id)
        else:
            raise RxDException(
                "Instruction for ParameterOnRegion semi_compile was neither 1d or 3d"
            )


class ParameterOnExtracellular(SpeciesOnExtracellular):
    pass


class State(Species):
    def __repr__(self):
        return "State(regions=%r, d=%r, name=%r, charge=%r, initial=%r)" % (
            self._regions,
            self._d,
            self._name,
            self._charge,
            self.initial,
        )



"""
Executes one simulation with specified stimulus.
Displays response and reports spike frequency
"""

from neuron import h, gui

# simulation parameter
h.tstop = 500  # ms, more than long enough for 15 spikes at ISI = 25 ms

# model specification
from cell import Cell

cell = Cell()

# instrumentation

# experimental manipulations
stim = h.IClamp(cell.soma(0.5))
stim.delay = 1  # ms
stim.dur = 1e9
stim.amp = 0.1  # nA

# data recording and analysis
# count only those spikes that get to distal end of dend
nc = h.NetCon(cell.dend(1)._ref_v, None, sec=cell.dend)
nc.threshold = -10  # mV
spvec = h.Vector()
nc.record(spvec)

NSETTLE = 5  # ignore the first NSETTLE ISI (allow freq to stablize)
NINVL = 10  # num ISI from which frequency will be calculated
NMIN = NSETTLE + NINVL  # ignore recordings with fewer than this num of ISIs


def get_frequency():
    nspikes = spvec.size()
    if nspikes > NMIN:
        t2 = spvec[-1]  # last spike
        t1 = spvec[-(1 + NINVL)]  # NINVL prior to last spike
        return NINVL * 1.0e3 / (t2 - t1)
    else:
        return 0


# Simulation control and reporting of results


def onerun(amp):
    # amp = amplitude of stimulus current
    g = h.Graph(0)
    g.size(0, 500, -80, 40)
    g.view(0, -80, 500, 120, 2, 105, 300.48, 200.32)
    # update graph throughout the simulation
    h.graphList[0].append(g)
    # plot v at distal end of dend
    g.addvar("dend(1).v", cell.dend(1)._ref_v)
    stim.amp = amp
    h.run()
    freq = get_frequency()
    print("stimulus %g frequency %g" % (amp, freq))


onerun(0.15)



from math import pi

import pytest

from testutils import compare_data, tol


@pytest.fixture
def ecs_example(neuron_instance):
    """A model where something is created in one cell and diffuses to another.

    This model makes use of parameters, multicompartment reactions, an NMODL
    and the extracellular space. A substance is created in an organelle in cell1
    and leaks into the cytosol. It then enters the ECS using an NMODL mechanism,
    where it diffuses to cell 2 and enters via the same NMODL mechanism.
    """

    h, rxd, data, save_path = neuron_instance

    def make_model(alpha, lambd):
        # create cell1 where `x` will be created and leak out
        cell1 = h.Section(name="cell1")
        cell1.pt3dclear()
        cell1.pt3dadd(-2, 0, 0, 1)
        cell1.pt3dadd(-1, 0, 0, 1)
        cell1.nseg = 11
        cell1.insert("pump")

        # create cell2 where `x` will be pumped in and accumulate
        cell2 = h.Section(name="cell2")
        cell2.pt3dclear()
        cell2.pt3dadd(1, 0, 0, 1)
        cell2.pt3dadd(2, 0, 0, 1)
        cell2.nseg = 11
        cell2.insert("pump")

        # Where?
        # the intracellular spaces
        cyt = rxd.Region(
            h.allsec(),
            name="cyt",
            nrn_region="i",
            geometry=rxd.FractionalVolume(0.9, surface_fraction=1.0),
        )

        org = rxd.Region(h.allsec(), name="org", geometry=rxd.FractionalVolume(0.1))

        cyt_org_membrane = rxd.Region(
            h.allsec(),
            name="mem",
            geometry=rxd.ScalableBorder(pi / 2.0, on_cell_surface=False),
        )

        # the extracellular space
        ecs = rxd.Extracellular(
            -55, -55, -55, 55, 55, 55, dx=33, volume_fraction=alpha, tortuosity=lambd
        )

        # Who?
        x = rxd.Species(
            [cyt, org, cyt_org_membrane, ecs], name="x", d=1.0, charge=1, initial=0
        )
        Xcyt = x[cyt]
        Xorg = x[org]

        # What? - produce X in cell 1
        # parameter to limit production to cell 1
        cell1_param = rxd.Parameter(
            org, initial=lambda node: 1.0 if node.segment.sec == cell1 else 0
        )

        # production with a rate following Michaels Menton kinetics
        createX = rxd.Rate(Xorg, cell1_param[org] * 1.0 / (10.0 + Xorg))

        # leak between organelles and cytosol
        cyt_org_leak = rxd.MultiCompartmentReaction(
            Xcyt, Xorg, 1e4, 1e4, membrane=cyt_org_membrane
        )
        model = (
            cell1,
            cell2,
            cyt,
            org,
            cyt_org_membrane,
            ecs,
            x,
            Xcyt,
            Xorg,
            createX,
            cell1_param,
            createX,
            cyt_org_leak,
        )
        return model

    yield (neuron_instance, make_model)


def test_ecs_example(ecs_example):
    """Test ecs_example with fixed step methods"""

    (h, rxd, data, save_path), make_model = ecs_example
    model = make_model(0.2, 1.6)
    h.finitialize(-65)
    h.continuerun(1000)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_example_cvode(ecs_example):
    """Test ecs_example with variable step methods"""

    (h, rxd, data, save_path), make_model = ecs_example
    model = make_model(0.2, 1.6)
    h.CVode().active(True)
    h.CVode().atol(1e-5)

    h.finitialize(-65)
    h.continuerun(1000)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_example_alpha(ecs_example):
    """Test ecs_example with fixed step and inhomogeneous volume fraction methods"""

    (h, rxd, data, save_path), make_model = ecs_example
    model = make_model(lambda x, y, z: 0.2, 1.6)
    h.finitialize(-65)
    h.continuerun(1000)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_example_cvode_alpha(ecs_example):
    """Test ecs_example with variable step and inhomogeneous volume fraction
    methods"""

    (h, rxd, data, save_path), make_model = ecs_example
    model = make_model(lambda x, y, z: 0.2, 1.6)
    h.CVode().active(True)
    h.CVode().atol(1e-5)

    h.finitialize(-65)
    h.continuerun(1000)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_example_tort(ecs_example):
    """Test ecs_example with fixed step and inhomogeneous tortuosity methods"""

    (h, rxd, data, save_path), make_model = ecs_example
    model = make_model(lambda x, y, z: 0.2, 1.6)
    h.finitialize(-65)
    h.continuerun(1000)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_example_cvode_tort(ecs_example):
    """Test ecs_example with variable step and inhomogeneous tortuosity methods"""

    (h, rxd, data, save_path), make_model = ecs_example
    model = make_model(lambda x, y, z: 0.2, 1.6)
    h.CVode().active(True)
    h.CVode().atol(1e-5)

    h.finitialize(-65)
    h.continuerun(1000)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_nodelists(ecs_example):
    """Test accessing species nodes with both Node1D and NodeExtracellular"""

    (h, rxd, data, save_path), make_model = ecs_example
    model = make_model(0.2, 1.6)
    ecs, x = model[5], model[6]
    # test accessing NodeExtracellular from species with both 1D and ECS
    assert len(x.nodes(ecs)) == 64
    assert all([nd.region == ecs for nd in x.nodes(ecs)])
    # test accessing specific node by location
    nd = x[ecs].nodes((-38, 27, 27))[0]
    assert (nd.x3d, nd.y3d, nd.z3d) == (-38.5, 27.5, 27.5)



from .rxdException import RxDException
from neuron import h

try:
    from . import geometry3d, dimension3
except:
    pass
import copy
import itertools
import numpy
from . import geometry as geo
import weakref
from . import initializer
import warnings
import math
import ctypes

_all_regions = []
_region_count = 0
_c_region_lookup = (
    None  # a global dictionary linking a region to a list of _c_regions it is a part of
)


def _sort_secs(secs):
    # sort the sections
    root_secs = h.SectionList()
    root_secs.allroots()
    all_sorted = h.SectionList()
    for root in root_secs:
        all_sorted.wholetree(sec=root)
    secs_names = dict([(sec.hoc_internal_name(), sec) for sec in secs])
    # for sec in secs:
    #    if sec.orientation():
    #        raise RxDException('still need to deal with backwards sections')
    secs = [
        secs_names[sec.hoc_internal_name()]
        for sec in all_sorted
        if sec.hoc_internal_name() in secs_names
    ]
    # return an empty list rather than an empty SectionList because
    # bool([]) == False
    # bool(h.SectionList([])) == True
    # secs are checked in some reactions to determine active regions
    return [] if secs == [] else h.SectionList(secs)


class _c_region:
    """
    The overlapping regions that are used to parse the relevant indices for JIT C reactions.
    regions - a set of regions that occur in the same sections
    """

    def __init__(self, regions):
        global _c_region_lookup
        self._regions = [weakref.ref(r) for r in regions]
        self._overlap = self._regions[0]()._secs1d
        self.num_regions = len(self._regions)
        self.num_species = 0
        self.num_params = 0
        self.num_ecs_species = 0
        self.num_ecs_params = 0
        self._ecs_react_species = list()
        self._ecs_react_params = list()
        self._react_species = list()
        self._react_params = list()
        self._react_regions = dict()
        self._initialized = False
        self.location_index = None
        self.ecs_location_index = None
        self._ecs_species_ids = None
        self._ecs_params_ids = None
        self._voltage_dependent = False
        self._vptrs = None
        for rptr in self._regions:
            r = rptr()
            self._overlap = h.SectionList(
                [sec for sec in r._secs1d if sec in self._overlap]
            )
            if r in _c_region_lookup:
                _c_region_lookup[rptr].append(self)
            else:
                _c_region_lookup[rptr] = [self]

    def add_reaction(self, rptr, region):
        # for multicompartment reaction -- check all regions are present
        if rptr() and hasattr(rptr(), "_changing_species"):
            for sptr in rptr()._changing_species:
                if (
                    sptr()
                    and hasattr(sptr(), "_region")
                    and sptr()._region not in self._regions
                ):
                    return
        if rptr in self._react_regions:
            if region not in self._react_regions[rptr]:
                self._react_regions[rptr].append(region)
        else:
            self._react_regions[rptr] = [region]
        if not self._voltage_dependent:
            self._voltage_dependent = rptr()._voltage_dependent
        self._initialized = False

    def add_species(self, species_set):
        from .species import SpeciesOnRegion, Parameter, ParameterOnRegion

        for s in species_set:
            if isinstance(s, ParameterOnRegion):
                if s._species() and s._species not in self._react_params:
                    self._react_params.append(s._species)
            elif isinstance(s, Parameter):
                if s._species not in self._react_params:
                    self._react_params.append(s._species)
            elif isinstance(s, SpeciesOnRegion):
                if s._species() and s._species not in self._react_species:
                    self._react_species.append(s._species)
            elif weakref.ref(s) not in self._react_species:
                self._react_species.append(weakref.ref(s))
        self.num_params = len(self._react_params)
        self.num_species = len(self._react_species)
        self._initilized = False

    def add_ecs_species(self, species_set):
        from .species import ParameterOnExtracellular

        for s in species_set:
            sptr = s._extracellular
            if isinstance(s, ParameterOnExtracellular):
                if sptr not in self._ecs_react_params:
                    self._ecs_react_params.append(sptr)
            else:
                if sptr not in self._ecs_react_species:
                    self._ecs_react_species.append(sptr)
        self.num_ecs_params = len(self._ecs_react_params)
        self.num_ecs_species = len(self._ecs_react_species)
        self._initialized = False

    def get_ecs_index(self):
        if not self._initialized:
            self._initalize()
        if self.ecs_location_index is None:
            return numpy.ndarray(0, ctypes.c_int)
        else:
            return self.ecs_location_index.flatten()

    def get_state_index(self):
        if not self._initialized:
            self._initalize()
        return self.location_index.flatten()

    def get_ecs_species_ids(self):
        ret = numpy.ndarray(self.num_ecs_species + self.num_ecs_params, ctypes.c_int)
        if self.num_ecs_species + self.num_ecs_params > 0:
            for i in self._ecs_species_ids:
                ret[self._ecs_species_ids[i]] = i
            for i in self._ecs_params_ids:
                ret[self._ecs_params_ids[i] + self.num_ecs_species] = i
        return ret

    def _ecs_initalize(self):
        from . import species

        self.ecs_location_index = -numpy.ones(
            (self.num_ecs_species + self.num_ecs_params, self.num_segments),
            ctypes.c_int,
        )

        self._ecs_react_species.sort(
            key=lambda sp: sp()._extracellular()._grid_id
            if isinstance(sp(), species.SpeciesOnExtracellular)
            else sp()._grid_id
        )
        self._ecs_react_params.sort(
            key=lambda sp: sp()._extracellular()._grid_id
            if isinstance(sp(), species.ParameterOnExtracellular)
            else sp()._grid_id
        )

        # Set the local ids of the regions and species involved in the reactions
        self._ecs_species_ids = dict()
        self._ecs_params_ids = dict()
        for sid, s in enumerate(self._ecs_react_species):
            self._ecs_species_ids[s()._grid_id] = sid
        for sid, s in enumerate(self._ecs_react_params):
            self._ecs_params_ids[s()._grid_id] = sid

        # Setup the matrix to the ECS grid points
        for sid, s in enumerate(self._ecs_react_species + self._ecs_react_params):
            seg_idx = 0
            for sec in self._overlap:
                for seg in sec:
                    (x, y, z) = species._xyz(seg)
                    self.ecs_location_index[sid][seg_idx] = s().index_from_xyz(x, y, z)
                    seg_idx += 1
        self.ecs_location_index = self.ecs_location_index.transpose()

    def _initalize(self):
        from .species import Species, SpeciesOnRegion

        self.num_segments = numpy.sum([x.nseg for x in self._overlap])
        self.location_index = -numpy.ones(
            (self.num_regions, self.num_species + self.num_params, self.num_segments),
            ctypes.c_int,
        )
        from .species import SpeciesOnExtracellular, SpeciesOnRegion, ParameterOnRegion

        # Set the local ids of the regions and species involved in the reactions
        self._species_ids = dict()
        self._params_ids = dict()
        self._region_ids = dict()
        self._react_species.sort(
            key=lambda sp: sp()._species()._id
            if isinstance(sp(), SpeciesOnRegion)
            else sp()._id
        )
        self._react_params.sort(
            key=lambda sp: sp()._species()._id
            if isinstance(sp(), ParameterOnRegion)
            else sp()._id
        )

        self._regions.sort(key=lambda rp: rp()._id)

        for rid, r in enumerate(self._regions):
            self._region_ids[r()._id] = rid
        for sid, s in enumerate(self._react_species):
            self._species_ids[s()._id] = sid
        for sid, s in enumerate(self._react_params):
            self._params_ids[s()._id] = sid

        # Setup the array to the state index
        for rid, r in enumerate(self._regions):
            for sid, s in enumerate(self._react_species + self._react_params):
                indices = s()._indices1d(r())
                if len(indices) > self.num_segments:
                    indices = s()._indices1d(r(), self._overlap)
                try:
                    if indices == []:
                        self.location_index[rid][sid][:] = -1
                    else:
                        self.location_index[rid][sid][:] = indices
                except ValueError:
                    raise RxDException(
                        "Rates and Reactions with species defined on different regions are not currently supported."
                    )
        self.location_index = self.location_index.transpose()

        if self._voltage_dependent:
            self._vptrs = []
            for sec in self._overlap:
                for seg in sec:
                    self._vptrs.append(seg._ref_v)

        # Setup the matrix to the ECS grid points
        if self.num_ecs_species + self.num_ecs_params > 0:
            self._ecs_initalize()
        self._initialized = True


class Extracellular:
    """Declare an extracellular region

    tortuosity = increase in path length due to obstacles, effective diffusion coefficient d/tortuosity^2. - either a single value for the whole region or a Vector giving a value for each voxel.

    Assumes reflective boundary conditions, so you'll probably want to make sure the mesh extends far beyond the neuron

    Assumes volume_fraction=1, i.e. that the extracellular space is empty. You'll probably want to lower this to
    account for other cells. Note: the section volumes are assumed to be negligible and are ignored by the simulation.

    Assumes tortuosity=1.
    """

    def __init__(
        self,
        xlo,
        ylo,
        zlo,
        xhi,
        yhi,
        zhi,
        dx,
        volume_fraction=1,
        tortuosity=None,
        permeability=None,
    ):
        from . import options

        if not options.enable.extracellular:
            raise RxDException(
                "Extracellular diffusion support is disabled. Override with rxd.options.enable.extracellular = True."
            )
        self._xlo, self._ylo, self._zlo = xlo, ylo, zlo
        if not hasattr(dx, "__len__"):
            self._dx = (dx, dx, dx)
        elif len(dx) == 3:
            self._dx = dx
        else:
            raise RxDException(
                "Extracellular region dx=%s is invalid, dx should be a number or a tuple (dx,dy,dz) for the length, width and height of the voxels"
                % repr(dx)
            )

        self._nx = int(math.ceil(float(xhi - xlo) / self._dx[0]))
        self._ny = int(math.ceil(float(yhi - ylo) / self._dx[1]))
        self._nz = int(math.ceil(float(zhi - zlo) / self._dx[2]))
        self._xhi, self._yhi, self._zhi = (
            xlo + float(self._dx[0]) * self._nx,
            ylo + float(self._dx[1]) * self._ny,
            zlo + float(self._dx[2]) * self._nz,
        )

        self._alpha, self._ecs_alpha = self._parse_volume_fraction(volume_fraction)

        if tortuosity is not None and permeability is not None:
            raise RxDException(
                "Specify either permeability or tortuosity, not both; permeability=1/tortuosity**2"
            )
        elif tortuosity is None and permeability is None:
            tortuosity = 1.0

        if permeability is None:
            self._tortuosity, self._ecs_permeability = self._parse_tortuosity(
                tortuosity
            )
        else:
            self._permeability, self._ecs_permeability = self._parse_tortuosity(
                permeability, True
            )

    def __repr__(self):
        return (
            "Extracellular(xlo=%r, ylo=%r, zlo=%r, xhi=%r, yhi=%r, zhi=%r, tortuosity=%r, volume_fraction=%r)"
            % (
                self._xlo,
                self._ylo,
                self._zlo,
                self._xhi,
                self._yhi,
                self._zhi,
                self.tortuosity,
                self.alpha,
            )
        )

    def _short_repr(self):
        return "Extracellular"

    def volume(self, index):
        """Returns the volume of the voxel at a given index"""
        if numpy.isscalar(self.alpha):
            return numpy.prod(self._dx) * self.alpha
        return numpy.prod(self._dx) * self.alpha[index]

    @property
    def _volume_fraction_vector(self):
        if self._ecs_alpha is None:
            self._ecs_alpha = self.alpha._extracellular_instances[self]._states
        return self._ecs_alpha

    @property
    def _permeability_vector(self):
        if self._ecs_permeability is None:
            self._ecs_permeability = self.permeability._extracellular_instances[
                self
            ]._states
        return self._ecs_permeability

    def _parse_tortuosity(self, value, is_permeability=False):
        if numpy.isscalar(value):
            parsed_value = value
            ecs_permeability = value if is_permeability else value ** (-2)
        elif callable(value):
            ecs_permeability = h.Vector()
            parsed_value = numpy.ndarray((self._nx, self._ny, self._nz))
            for i in range(self._nx):
                for j in range(self._ny):
                    for k in range(self._nz):
                        parsed_value[i, j, k] = value(
                            self._xlo + i * self._dx[0],
                            self._ylo + j * self._dx[1],
                            self._zlo + k * self._dx[2],
                        )
            if is_permeability:
                ecs_permeability = h.Vector(parsed_value.flatten())
            else:
                ecs_permeability = h.Vector(parsed_value.flatten()).pow(-2)
        elif hasattr(value, "_extracellular_instances") or hasattr(
            self, "_extracellular"
        ):
            # Check Species or SpeciesOnExtracellular is defined on this region
            if (
                hasattr(value, "_extracellular_instances")
                and self not in value._extracellular_instances
            ) or (
                (
                    hasattr(value, "_extracellular")
                    and value._extracellular()
                    and value._extracellular() != self
                )
            ):
                raise RxDException(
                    "permeability can be set to a State or Parameter, but it must be defined on this Extracellular region %r"
                    % self
                )
            else:
                if is_permeability:
                    parsed_value = value
                    ecs_permeability = None
                else:
                    raise RxDException(
                        "Only permeability can be set to a State or Parameter. The tortuosity must be a; scalar, function (of x,y,z locations), or array the size of the extracellular space"
                    )
        elif isinstance(value, dict):
            if is_permeability:
                from .species import State

                params = value.copy()
                params["regions"] = [self]
                parsed_value = State(**params)
                ecs_permeability = None
            else:
                raise RxDException(
                    "Only permeability can be set to a State or Parameter. The tortuosity must be a; scalar, function (of x,y,z locations), or array the size of the extracellular space"
                )

        else:
            parsed_value = numpy.array(value)
            if parsed_value.shape != (self._nx, self._ny, self._nz):
                raise RxDException(
                    "permeability or tortuosity must be a scalar, a function of the (x,y,z) location or an array the same size as the grid: {0}x{1}x{2}".format(
                        self._nx, self._ny, self._nz
                    )
                )

            else:
                if is_permeability:
                    ecs_permeability = h.Vector(parsed_value.flatten())
                else:
                    ecs_permeability = h.Vector(parsed_value.flatten()).pow(-2)
        return parsed_value, ecs_permeability

    def _parse_volume_fraction(self, volume_fraction):

        if numpy.isscalar(volume_fraction):
            alpha = float(volume_fraction)
            alpha = alpha
            ecs_alpha = alpha
        elif isinstance(volume_fraction, dict):
            from .species import State

            params = volume_fraction.copy()
            params["regions"] = [self]
            alpha = State(**params)
            ecs_alpha = None
            warnings.warn(
                "Dynamic changes to the volume fraction does not change the concentrations of species already in the region."
            )
        elif hasattr(volume_fraction, "_extracellular_instances") or hasattr(
            self, "_extracellular"
        ):
            # Check Species or SpeciesOnExtracellular is defined on this region
            if (
                hasattr(volume_fraction, "_extracellular_instances")
                and self not in volume_fraction._extracellular_instances
            ) or (
                (
                    hasattr(volume_fraction, "_extracellular")
                    and volume_fraction._extracellular()
                    and volume_fraction._extracellular() != self
                )
            ):
                raise RxDException(
                    "volume fraction can be set to a State or Parameter, but it must be defined on this Extracellular region %r"
                    % self
                )
            else:
                alpha = volume_fraction
                ecs_alpha = None
                warnings.warn(
                    "Dynamic changes to the volume fraction does not change the concentrations of species already in the region."
                )

        elif callable(volume_fraction):
            alpha = numpy.ndarray((self._nx, self._ny, self._nz))
            for i in range(self._nx):
                for j in range(self._ny):
                    for k in range(self._nz):
                        alpha[i, j, k] = volume_fraction(
                            self._xlo + i * self._dx[0],
                            self._ylo + j * self._dx[1],
                            self._zlo + k * self._dx[2],
                        )
                alpha = alpha
                ecs_alpha = h.Vector(alpha.flatten())
        else:
            alpha = numpy.array(volume_fraction)
            if alpha.shape != (self._nx, self._ny, self._nz):
                raise RxDException(
                    "free volume fraction alpha must be a scalar or an array the same size as the grid: {0}x{1}x{2}".format(
                        self._nx, self._ny, self._nz
                    )
                )

            else:
                alpha = h.Vector(alpha)
                ecs_alpha = self._alpha.as_numpy().reshape(self._nx, self._ny, self._nz)
        return alpha, ecs_alpha

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        """Set the value of the volume fraction for all species define on
           this Extracellular region. Changing the volume fraction will not
           change the concentrations already there, it will alter the
           magnitude of concentration changes due to currents.

        Args:
            value (float, array, callable or State) the new volume fraction, it
                can be a scalar for homogeneous porosity, a array the size of
                the extracellular space, a function that take (x,y,z) as an
                argument and will be evaluated on every extracellular voxel, or
                a State defined on the same extracellular region.
        """

        from .species import _update_volume_fraction

        self._alpha, self._ecs_alpha = self._parse_volume_fraction(value)
        _update_volume_fraction(self)

    @property
    def tortuosity(self):
        if hasattr(self, "_tortuosity"):
            return self._tortuosity
        return (1.0 / self._permeability) ** 0.5

    @property
    def permeability(self):
        if hasattr(self, "_tortuosity"):
            return 1.0 / self._tortuosity**2
        return self._permeability

    @tortuosity.setter
    def tortuosity(self, value):
        """Set the value of the tortuosity for all species define on this
           Extracellular region.

        Args:
            value (float, array or callable) the new tortuosity, it can be a
                scalar for homogeneous diffusion, or a array the size of the
                extracellular space or a function that take (x,y,z) as an
                argument and will be evaluated on every extracellular voxel.
        """

        from .species import _update_tortuosity

        self._tortuosity, self._ecs_permeability = self._parse_tortuosity(value)
        if hasattr(self, "_permeability"):
            delattr(self, "_permeability")

        # update tortuosity for any species defined on this ECS
        _update_tortuosity(self)

    @permeability.setter
    def permeability(self, value):
        """Set the value of the permeability for all species define on this
           Extracellular region.

        Args:
            value (float, array or callable) the new tortuosity, it can be a
                scalar for homogeneous diffusion, or a array the size of the
                extracellular space or a function that take (x,y,z) as an
                argument and will be evaluated on every extracellular voxel, or
                an rxd State or Parameter defined on the extracellular region.
        """

        from .species import _update_tortuosity

        self._permeability, self._ecs_permeability = self._parse_tortuosity(value, True)
        if hasattr(self, "_tortuosity"):
            delattr(self, "_tortuosity")

        # update tortuosity for any species defined on this ECS
        _update_tortuosity(self)


class Region(object):
    """Declare a conceptual region of the neuron.

    Examples: Cytosol, ER
    """

    def __repr__(self):
        # Note: this used to print out dimension, but that's now on a per-segment basis
        # TODO: remove the note when that is fully true
        return "Region(..., nrn_region=%r, geometry=%r, dx=%r, name=%r)" % (
            self.nrn_region,
            self._geometry,
            self.dx,
            self._name,
        )

    def __contains__(self, item):
        try:
            if item.region == self:
                return True
            else:
                return False
        except:
            raise NotImplementedError()

    def _short_repr(self):
        if self._name is not None:
            return str(self._name)
        else:
            return self.__repr__()

    def mesh_eval(self):
        from .species import _all_species

        for spref in _all_species:
            sp = spref()
            if sp is not None:
                if self in sp.regions:
                    break
        else:
            return {"instantiated": False}
        result = {"instantiated": True}
        total_num_3d_segs = sum(sec.nseg for sec in self._secs3d)
        segs_with_surface = {node.segment for node in sp.nodes if node.surface_area}
        result["3dsegswithoutsurface"] = len(segs_with_surface) - total_num_3d_segs
        return result

    def _do_init(self):
        global _region_count

        # _do_init can be called multiple times due to a change in geometry
        if hasattr(self, "_allow_setting"):
            # here are things that must only happen once
            del self._allow_setting
            self._id = _region_count
            _region_count += 1

        from . import rxd

        # parameters that were defined in old init
        # TODO: remove need for this bit
        nrn_region = self.nrn_region

        # TODO: self.dx needs to be removed... eventually that should be on a per-section basis
        #       right now, it has to be consistent but 2this is unenforced
        if self.dx is None:
            self.dx = 0.25
        dx = self.dx

        self._secs1d = []
        self._secs3d = []

        dims = rxd._domain_lookup
        for sec in self._secs:
            dim = dims(sec)
            if dim == 1:
                self._secs1d.append(sec)
            elif dim == 3:
                self._secs3d.append(sec)
            else:
                raise RxDException(
                    "unknown dimension: %r in section %r" % (dim, sec.name())
                )

        # TODO: I used to not sort secs in 3D if hasattr(self._secs, 'sections'); figure out why
        self._secs = _sort_secs(self._secs)
        self._secs1d = _sort_secs(self._secs1d)
        self._secs3d = h.SectionList(self._secs3d)

        if any(self._secs3d):
            if not (hasattr(self._geometry, "volumes3d")):
                raise RxDException(
                    'selected geometry (%r) does not support 3d mode (no "volumes3d" attr)'
                    % self._geometry
                )

            if nrn_region == "o":
                raise RxDException('3d simulations do not support nrn_region="o" yet')

            if hasattr(self, "_voxelization"):
                internal_voxels, surface_voxels, mesh_grid = (
                    self._voxelization["internal_voxels"],
                    self._voxelization["surface_voxels"],
                    self._voxelization["mesh_grid"],
                )
            else:
                internal_voxels, surface_voxels, mesh_grid = self._geometry.volumes3d(
                    self._secs3d, dx=dx
                )

            self._sa = numpy.zeros(len(surface_voxels) + len(internal_voxels))
            self._vol = numpy.ones(len(surface_voxels) + len(internal_voxels))
            self._mesh_grid = mesh_grid
            self._points = [key for key in surface_voxels.keys()] + [
                key for key in internal_voxels.keys()
            ]

            self._points = sorted(self._points)

            nodes_by_seg = {}
            surface_nodes_by_seg = {}
            # creates tuples of x, y, and z coordinates where a point is (xs[i], ys[i], zs[i])
            self._xs, self._ys, self._zs = zip(*self._points)
            maps = {
                seg: i
                for i, seg in enumerate([seg for sec in self._secs3d for seg in sec])
            }
            segs = []

            for i, p in enumerate(self._points):
                if p in surface_voxels:
                    vx = surface_voxels[p]
                    self._vol[i], self._sa[i], idx = vx[0], vx[1], maps[vx[2]]
                    segs.append(idx)

                    surface_nodes_by_seg.setdefault(idx, [])
                    surface_nodes_by_seg[idx].append(i)

                    nodes_by_seg.setdefault(idx, [])
                    nodes_by_seg[idx].append(i)
                    self._sa[i] = surface_voxels[p][1]

                else:
                    vx = internal_voxels[p]
                    self._vol[i], idx = vx[0], maps[vx[1]]
                    segs.append(idx)

                    nodes_by_seg.setdefault(idx, [])
                    nodes_by_seg[idx].append(i)

            self._surface_nodes_by_seg = surface_nodes_by_seg
            self._nodes_by_seg = nodes_by_seg
            self._secs3d_names = {
                sec.hoc_internal_name(): sec.nseg for sec in self._secs3d
            }
            self._segsidx = segs
            self._dx = self.dx

    def _segs3d(self, index=None):
        segs = [seg for sec in self._secs3d for seg in sec]
        if index:
            return [seg for i, seg in enumerate(segs) if i in index]
        return segs

    def _indices_from_sec_x(self, sec, position):
        # TODO: the assert is here because the diameter is not computed correctly
        #       unless it coincides with a 3d point, which we only know to exist at the
        #       endpoints and because the section does not proceed linearly between
        #       the endpoints (in general)... which affects the computation of the
        #       normal vector as well
        assert position in (0, 1)
        # NOTE: some care is necessary in constructing normal vector... must be
        #       based on end frusta, not on vector between end points
        dx = self.dx
        if position == 0:
            x = sec.x3d(0)
            y = sec.y3d(0)
            z = sec.z3d(0)
            nx = sec.x3d(1) - x
            ny = sec.y3d(1) - y
            nz = sec.z3d(1) - z
        elif position == 1:
            n = sec.n3d()
            x = sec.x3d(n - 1)
            y = sec.y3d(n - 1)
            z = sec.z3d(n - 1)
            nx = x - sec.x3d(n - 2)
            ny = y - sec.y3d(n - 2)
            nz = z - sec.z3d(n - 2)
            x -= dx * nx / (nx**2 + ny**2 + nz**2) ** 0.5
            y -= dx * ny / (nx**2 + ny**2 + nz**2) ** 0.5
            z -= dx * nz / (nx**2 + ny**2 + nz**2) ** 0.5

        else:
            raise RxDException("should never get here")
        if nx == 0 and ny == 0 and nz == 0:
            raise RxDException(
                "The 3D/1D join on section %r cannot be calculated from identical 3d points %g, %g, %g"
                % (sec, x, y, z)
            )
        # dn = (nx**2 + ny**2 + nz**2)**0.5
        # nx, ny, nz = nx/dn, ny/dn, nz/dn
        # x, y, z = x * x1 + (1 - x) * x0, x * y1 + (1 - x) * y0, x * z1 + (1 - x) * z1
        r = sec(position).diam * 0.5 + self.dx * 3**0.5
        plane_of_disc = geometry3d.graphicsPrimitives.Plane(x, y, z, nx, ny, nz)
        potential_coordinates = []

        xs = numpy.arange(
            self._mesh_grid["xlo"] - self._mesh_grid["dx"],
            self._mesh_grid["xhi"] + self._mesh_grid["dx"],
            self._mesh_grid["dx"],
        )
        ys = numpy.arange(
            self._mesh_grid["ylo"] - self._mesh_grid["dy"],
            self._mesh_grid["yhi"] + self._mesh_grid["dy"],
            self._mesh_grid["dy"],
        )
        zs = numpy.arange(
            self._mesh_grid["zlo"] - self._mesh_grid["dz"],
            self._mesh_grid["zhi"] + self._mesh_grid["dz"],
            self._mesh_grid["dz"],
        )
        xlo = self._mesh_grid["xlo"] - self._mesh_grid["dx"]
        ylo = self._mesh_grid["ylo"] - self._mesh_grid["dy"]
        zlo = self._mesh_grid["zlo"] - self._mesh_grid["dz"]
        # locate the indices of the cube containing the sphere containing the disc
        # TODO: write this more efficiently
        i_indices = [i for i, a in enumerate(xs) if abs(a - x) <= r]
        j_indices = [i for i, a in enumerate(ys) if abs(a - y) <= r]
        k_indices = [i for i, a in enumerate(zs) if abs(a - z) <= r]
        sphere_indices = [
            (i, j, k)
            for i, j, k in itertools.product(i_indices, j_indices, k_indices)
            if (xs[i] - x) ** 2 + (ys[j] - y) ** 2 + (zs[k] - z) ** 2 <= r**2
        ]
        dx = self.dx
        disc_indices = []
        for i, j, k in sphere_indices:
            #            a, b, c = xs[i], ys[j], zs[k]
            # TODO: no need to compute all; can stop when some True and some False
            #            on_side1 = [plane_of_disc.distance(x, y, z) >= 0 for x, y, z in itertools.product([a - dx2, a + dx2], [b - dx2, b + dx2], [c - dx2, c + dx2])]
            # NOTE: the expression is structured this way to make sure it tests the exact same corner coordinates for corners shared by multiple voxels and that there are no round-off issues (an earlier attempt had round-off issues that resulted in double-thick discs when the frustum ended exactly on a grid plane)
            on_side1 = [
                plane_of_disc.distance(x, y, z) > 0
                for x, y, z in itertools.product(
                    [xlo + (i - 0.5) * dx, xlo + (i + 0.5) * dx],
                    [ylo + (j - 0.5) * dx, ylo + (j + 0.5) * dx],
                    [zlo + (k - 0.5) * dx, zlo + (k + 0.5) * dx],
                )
            ]
            # need both sides to have at least one corner
            if any(on_side1) and not all(on_side1):
                # if we're here, then we've found a point on the disc.
                disc_indices.append((i - 1, j - 1, k - 1))
        return disc_indices

    def __init__(
        self,
        secs=None,
        nrn_region=None,
        geometry=None,
        dimension=None,
        dx=None,
        name=None,
    ):
        """
        In NEURON 7.4+, secs is optional at initial region declaration, but it
        must be specified before the reaction-diffusion model is instantiated.

        .. note:: dimension and dx will be deprecated in a future version
        """
        self._allow_setting = True
        if hasattr(secs, "__len__"):
            self._secs = secs
        else:
            self._secs = [secs]
        if secs == [] or secs is None:
            warnings.warn(
                "Warning: No sections. Region 'secs' should be a list of NEURON sections."
            )
        from nrn import Section

        for sec in self._secs:
            if not isinstance(sec, Section):
                raise RxDException(
                    "Error: Region 'secs' must be a list of NEURON sections, %r is not a valid NEURON section."
                    % sec
                )
        self._secs = h.SectionList(self._secs)
        self.nrn_region = nrn_region
        self.geometry = geometry

        if dimension is not None:
            warnings.warn(
                "dimension argument was a development feature only; use set_solve_type instead... the current version sets all the sections to your requested dimension, but this will override any previous settings"
            )
            import neuron

            neuron.rxd.set_solve_type(secs, dimension=dimension)
        self._name = name
        if dx is not None:
            try:
                dx = float(dx)
            except:
                dx = -1
            if dx <= 0:
                raise RxDException("dx must be a positive real number or None")
        self.dx = dx
        _all_regions.append(weakref.ref(self))

        # initialize self if the rest of rxd is already initialized
        if initializer.is_initialized():
            self._do_init()

    @property
    def nrn_region(self):
        """Get or set the classic NEURON region associated with this object.

        There are three possible values:
            * `'i'` -- just inside the plasma membrane
            * `'o'` -- just outside the plasma membrane
            * `None` -- none of the above

        .. note:: Setting only supported in NEURON 7.4+, and then only before the reaction-diffusion model is instantiated.
        """
        return self._nrn_region

    @nrn_region.setter
    def nrn_region(self, value):
        if hasattr(self, "_allow_setting"):
            if value not in (None, "i", "o"):
                raise RxDException('nrn_region must be one of: None, "i", "o"')
            else:
                self._nrn_region = value
        else:
            raise RxDException("Cannot set nrn_region now; model already instantiated")

    @property
    def geometry(self):
        """Get or set the geometry associated with this region.

        Setting the geometry to `None` will cause it to default to `rxd.geometry.inside`.

        .. note:: New in NEURON 7.4+. Setting allowed only before the reaction-diffusion model is instantiated.
        """
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        if hasattr(self, "_allow_setting"):
            if value is None:
                value = geo.inside
            self._geometry = value
        else:
            raise RxDException("Cannot set geometry now; model already instantiated")

    @property
    def name(self):
        """Get or set the Region's name.

        .. note:: New in NEURON 7.4+.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def _semi_compile(self):
        return "r%d" % self._id

    @property
    def secs(self):
        """Get or set the sections associated with this region.

        The sections may be expressed as a NEURON SectionList or as any Python
        iterable of sections.

        Note: The return value is a copy of the internal section list; modifying
              it will not change the Region.

        .. note:: Setting is new in NEURON 7.4+ and allowed only before the reaction-diffusion model is instantiated.
        """
        if hasattr(self._secs, "__len__"):
            return list(self._secs)
        else:
            return [self._secs]

    @secs.setter
    def secs(self, value):
        if hasattr(self, "_allow_setting"):
            if hasattr(value, "__len__"):
                secs = value
            else:
                secs = [value]

            from nrn import Section

            for sec in secs:
                if not isinstance(sec, Section):
                    raise RxDException(
                        "Error: Region 'secs' must be a list of NEURON sections, %r is not a valid NEURON section."
                        % sec
                    )
            self._secs = h.SectionList(secs)
        else:
            raise RxDException("Cannot set secs now; model already instantiated")

    def volume(self, index):
        """Returns the volume of the voxel at a given index"""
        return self._vol[index]



import pytest

from testutils import compare_data, tol


@pytest.fixture
def multicompartment_alignment(neuron_instance):
    """A model where the group Regions used for reactions result in too many
    1D indices found for a MulticompartmentReaction with the ECS.
    """

    h, rxd, data, save_path = neuron_instance
    soma = h.Section(name="soma")
    soma.L = soma.diam = 25
    soma.nseg = 1
    soma.pt3dclear()
    soma.pt3dadd(-12.5, 0, 0, 5)
    soma.pt3dadd(-7.5, 0, 0, 5)

    dend = h.Section(name="dend")
    dend.nseg = 101
    dend.pt3dclear()
    dend.pt3dadd(-7.5, 0, 0, 1)
    dend.pt3dadd(12.5, 0, 0, 1)
    dend.connect(soma)

    # Model parameters
    Dk = 2.62  # um^2/ms
    dx = 5  # um ECS voxel size
    alpha = 0.2  # ECS free volume fraction
    tortuosity = 1.6  # ECS tortuosity
    Dkeff = Dk / tortuosity**2  # um^2/ms effective diffusion coeffishent

    # Where -- shells and border between them
    # cytosol
    cyt = rxd.Region(h.allsec(), nrn_region="i", name="cyt")

    # cell membrane -- between cyt and FHS
    mem = rxd.Region(h.allsec(), name="mem", geometry=rxd.membrane())

    # Frankenhaeuser-Hodgkin space -- a shell with diameter 1.2 times that of
    #                                 the section
    fhs = rxd.Region([soma], name="fhs", nrn_region="o", geometry=rxd.Shell(1, 1.2))

    # a border -- between FHS and ECS
    bor = rxd.Region([soma], name="bor", geometry=rxd.ScalableBorder(diam_scale=1.2))

    # the shared ECS
    ecs = rxd.Extracellular(
        -15, -15, -15, 15, 15, 15, dx=dx, tortuosity=1.6, volume_fraction=alpha
    )

    # Who -- Potassium
    # scale factor so the flux (Dk/dx/2)*K has units molecules/um^2/ms
    # where dx/2 is the average distance between the center of the ECS voxel
    # and the FHS
    mM_to_mol_per_um = 6.0221409e23 * 1e-18

    # k in the cyt and FHS will link to the ion representation in NEURON, i.e.
    # seg.ki and seg.ko
    k = rxd.Species(
        [cyt, fhs],
        d=Dk,
        name="k",
        charge=1,
        initial=lambda nd: 140.0 if nd.region == cyt else 3.0,
    )

    # kecs is a fake extracellular potassium, it will only changed by rxd
    # mechanisms
    kecs = rxd.Species([ecs], d=Dk, name="kecs", charge=1, initial=3.0)

    # What -- use reactions to setup diffusion
    # Here we assume the FHS free volume fraction and tortuoisty are 1 and the
    # the thickness of the shell is much smaller than the ECS voxel size.
    diffusion = rxd.MultiCompartmentReaction(
        k[fhs],
        kecs[ecs],
        mM_to_mol_per_um * (Dkeff * alpha / (dx / 2)),
        mM_to_mol_per_um * (Dkeff * alpha / (dx / 2)),
        membrane=bor,
    )
    model = (soma, dend, cyt, fhs, bor, k, kecs, diffusion)
    yield (neuron_instance, model)


def test_ecs_multicompartment_alignment(multicompartment_alignment):
    """Test ecs_multicompartment_alignment model"""

    neuron_instance, model = multicompartment_alignment
    h, rxd, data, save_path = neuron_instance
    h.finitialize(-70)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



# -*- coding: utf-8 -*-
""" unit definitions """

# NEURON's default units
mM = 1
ms = 1
mV = 1
µm = um = 1

# concentration
µM = uM = 1e-3 * mM
nM = 1e-6 * mM
M = 1e3 * mM

# time
µs = us = 1e-3 * ms
s = sec = 1e3 * ms
minute = 60 * sec
hour = 60 * minute
day = 24 * hour

# space
nm = 1e-3 * um
mm = 1000 * um
cm = 10 * mm
m = 100 * cm

# voltage
μV = uV = 1e-3 * mV
V = 1e3 * mV



# Test of data return covering most of the functionality.
import distutils.util
import itertools
import os

from neuron import h, gui

pc = h.ParallelContext()
h.dt = 1.0 / 32
cvode = h.CVode()


class Cell:
    def __init__(self, gid):
        nsec = 5
        self.gid = gid
        r = h.Random()
        self.r = r
        r.Random123(gid, 0, 0)
        self.secs = [h.Section(name="s%d" % i, cell=self) for i in range(nsec)]
        s0 = self.secs[0]

        # random connect to exercise permute
        for i, s in enumerate(self.secs[1:]):
            s.connect(self.secs[int(r.discunif(0, i))])

        # hh and pas everywhere with random gkbar_hh and g_pas for intrinsic firing.
        # everywhere but s0.
        s = self.secs[0]
        s.L = 10
        s.diam = 10
        s.insert("hh")
        for s in self.secs[1:]:
            s.nseg = 4
            s.L = 50
            s.diam = 1
            s.insert("pas")
            s.e_pas = -65
            s.g_pas = r.uniform(0.0001, 0.0002)
            s.insert("hh")
            s.gkbar_hh = r.uniform(0.01, 0.02)

        pc.set_gid2node(gid, pc.id())
        pc.cell(gid, h.NetCon(s0(0.5)._ref_v, None, sec=s0))

        # add a few random Netstim -> NetCon -> Exp2Syn connections to verify
        # correct data return for ARTIFICIAL_CELL and POINT_PROCESS
        nsyn = 3
        self.stims = [h.NetStim() for _ in range(nsyn)]
        self.syns = [
            h.Exp2Syn(self.secs[int(r.discunif(0, nsec - 1))](0.5)) for _ in range(nsyn)
        ]
        self.ncs = [h.NetCon(self.stims[i], self.syns[i]) for i in range(nsyn)]
        for stim in self.stims:
            stim.start = int(r.uniform(0, 1) / h.dt) * h.dt
            stim.interval = int(r.uniform(1, 2) / h.dt) * h.dt
        for nc in self.ncs:
            nc.weight[0] = r.uniform(0.0, 0.001)
            nc.delay = int(r.discunif(5, 20)) * h.dt

    def __str__(self):
        return "Cell%d" % self.gid


class Network:
    def __init__(self):
        self.cells = [Cell(i) for i in range(5)]
        cvode.use_fast_imem(True)
        # a few intrinsically firing ARTIFICIAL_CELLS with and without gids
        self.acells = [h.IntervalFire() for _ in range(8)]
        r = h.Random()
        r.Random123(6, 0, 0)
        for a in self.acells:
            a.tau = r.uniform(2, 5)
            a.invl = r.uniform(2, 4)
        for i, a in enumerate(self.acells[5:]):
            pc.set_gid2node(i + 5, pc.id())
            pc.cell(i + 5, h.NetCon(a, None))

    def add_data(self, mname, d):
        instances = h.List(mname)
        if instances.count() == 0:
            return
        names = []
        inst = instances.o(0)
        for name in dir(inst):
            if "__" not in name:
                try:
                    if type(getattr(inst, name)) == float:
                        names.append(name)
                except:
                    pass
        for inst in instances:
            for name in names:
                d.append(((mname, name), getattr(inst, name)))

    def data(self):
        d = [("t", h.t)]
        for ncell, cell in enumerate(self.cells):
            for nsec, sec in enumerate(cell.secs):
                for nseg, seg in enumerate(sec):
                    d.append(((ncell, nsec, nseg, "v"), seg.v))
                    d.append(((ncell, nsec, nseg, "i"), seg.i_membrane_))
                    for mech in seg:
                        for var in mech:
                            d.append(
                                ((ncell, nsec, nseg, mech.name(), var.name()), var[0])
                            )

        # all NetStim, ExpSyn, ...
        for mname in ["NetStim", "Exp2Syn", "IntervalFire"]:
            self.add_data(mname, d)

        return d


show = False


def mkgraphs(model):
    h.newPlotV()
    gcell = h.graphList[0].o(h.graphList[0].count() - 1)
    gcell.erase_all()
    for cell in model.cells:
        for sec in cell.secs:
            gcell.addvar(" ", sec(0.5)._ref_v)

    h.newPlotV()
    gacell = h.graphList[0].o(h.graphList[0].count() - 1)
    gacell.erase_all()
    gacell.size(0, h.tstop, 0, h.tstop)
    for acell in model.acells:
        gacell.addvar(" ", acell._ref_t0)
    model.gui = (gcell, gacell)


def test_datareturn():
    from neuron import coreneuron

    coreneuron.enable = False

    model = Network()

    tstop = 5

    pc.set_maxstep(10)

    def run(tstop, mode):
        h.finitialize(-65)
        if mode == 0:
            pc.psolve(tstop)
        elif mode == 1:
            while h.t < tstop:
                pc.psolve(h.t + 1.0)
        else:
            while h.t < tstop:
                h.continuerun(h.t + 0.5)
                pc.psolve(h.t + 0.5)

    if show:
        # visual verify that cells and acells have different final results so that
        # data return comparisons are substantive.
        from neuron import gui

        h.steps_per_ms = 1.0 / h.dt
        h.setdt()
        h.tstop = tstop
        mkgraphs(model)
        h.run()
    else:
        run(tstop, 0)  # NEURON run
    std = model.data()

    print("CoreNEURON run")
    h.CVode().cache_efficient(1)
    coreneuron.enable = True
    coreneuron.verbose = 0
    coreneuron.gpu = bool(
        distutils.util.strtobool(os.environ.get("CORENRN_ENABLE_GPU", "false"))
    )

    results = []
    cell_permute_values = (1, 2) if coreneuron.gpu else (0, 1)
    for mode, nthread, cell_permute in itertools.product(
        [0, 1, 2], [1, 2], cell_permute_values
    ):
        pc.nthread(nthread)
        coreneuron.cell_permute = cell_permute
        run(tstop, mode)
        tst = model.data()
        max_diff = (
            h.Vector(v for k, v in std).sub(h.Vector(v for k, v in tst)).abs().max()
        )
        assert len(std) == len(tst)
        print(
            "mode={} nthread={} permute={} max diff={}".format(
                mode, nthread, cell_permute, max_diff
            )
        )
        results.append(max_diff)
        if max_diff > 1e-10:
            for i in range(len(std)):
                nrn_key, nrn_val = std[i]
                tst_key, tst_val = tst[i]
                assert nrn_key == tst_key
                if abs(nrn_val - tst_val) >= 1e-10:
                    print(i, nrn_key, nrn_val, tst_val, abs(nrn_val - tst_val))

    assert all(max_diff < 1e-10 for max_diff in results)

    if __name__ != "__main__":
        # tear down
        coreneuron.enable = False
        pc.nthread(1)
        pc.gid_clear()


if __name__ == "__main__":
    show = False
    test_datareturn()
    h.quit()



import sys
import re

voidfun = []
intvar = [[], []]  # scalarint vectorint
fltvar = [[], []]  # scalarfloat, vectorfloat
dblvar = [[], [], [], []]  # scalar, vectordouble, arraydouble, threedim


def processfun(a, names):
    # each name has trailing '()'
    for i in names:
        a.append(i.strip("()"))


def processvar(a, names):
    # each name may have 0-3 trailing [%d]
    for i in names:
        sp = i.split("[")
        b = []
        for j in sp:
            b.append(j.strip("]"))
        a[len(b) - 1].append(b)


def remove_multiline_comments(string):
    # remove all occurance comments (/*COMMENT */) from string
    string = re.sub(re.compile("/\*.*?\*/", re.DOTALL), "", string)
    return string


types = {}
types["void"] = (voidfun, processfun)
types["int"] = (intvar, processvar)
types["float"] = (fltvar, processvar)
types["double"] = (dblvar, processvar)


def process(type, names):
    types[type][1](types[type][0], names)


# read preprocessor output file into buffer
text = sys.stdin.read()

# the PGI preprocessor output i.e. 'pgcc -Mcpp -E'
# has multi-line comments (unless we specify -Mcpp=nocomment).
# we need to remove these multi-line comments
text = remove_multiline_comments(text)

# the pgcc 18.4 compiler prepends with a multiline typedef and several
# extern void lines that need to be skipped. Our first relevant line
# contains spatial_method or neuron2nemo or node_data

skip = 1
for line in text.splitlines():
    if "spatial_method" in line or "neuron2nemo" in line or "node_data" in line:
        skip = 0
    names = line.replace(",", " ").replace(";", " ").split()
    if not skip and len(names) > 2:
        print(line)  # entire filtered neuron.h file without pgcc added lines.
        process(names[1], names[2:])


print(
    """
/* Functions */
static VoidFunc functions[] = {
"""
)

for i in voidfun:
    if i:
        prefix = "nrnhoc_"
        j = i[len(prefix) :] if i.startswith(prefix) else i
        print('"%s", %s,' % (j, i))
print(
    """0, 0
};

static struct {  /* Integer Scalars */
  const char *name;
  int  *pint;
} scint[] = {
"""
)

for i in intvar[0]:
    print('"%s", &%s,' % (i[0], i[0]))
print(
    """0, 0
};

static struct {  /* Vector integers */
  const char *name;
  int  *pint;
  int  index1;
} vint[] = {
"""
)

for i in intvar[1]:
    print('"%s", %s, %s,' % (i[0], i[0], i[1]))
print(
    """0,0
};

static struct {  /* Float Scalars */
  const char *name;
  float  *pfloat;
} scfloat[] = {
"""
)

for i in fltvar[0]:
    print('"%s", &%s,' % (i[0], i[0]))
print(
    """0, 0
};

static struct {  /* Vector float */
  const char *name;
  float *pfloat;
  int index1;
} vfloat[] = {
"""
)

for i in fltvar[1]:
    print('"%s", %s, %s,' % (i[0], i[0], i[1]))
print(
    """0,0,0
};

/* Double Scalars */
DoubScal scdoub[] = {
"""
)

for i in dblvar[0]:
    print('"%s", &%s,' % (i[0], i[0]))
print(
    """0,0
};

/* Vectors */
DoubVec vdoub[] = {
"""
)

for i in dblvar[1]:
    print('"%s", %s, %s,' % (i[0], i[0], i[1]))
print(
    """0, 0, 0
};

static struct {  /* Arrays */
  const char *name;
  double *pdoub;
  int index1;
  int index2;
} ardoub[] = {
"""
)

for i in dblvar[2]:
    print('"%s", %s, %s, %s,' % (i[0], i[0], i[1], i[2]))
print(
    """0, 0, 0, 0
};

static struct {  /* triple dimensioned arrays */
  const char *name;
  double *pdoub;
  int index1;
  int index2;
  int index3;
} thredim[] = {
"""
)

for i in dblvar[3]:
    print('"%s", %s, %s, %s, %s,' % (i[0], i[0], i[1], i[2], i[3]))
print(
    """0, 0, 0, 0, 0
};
"""
)



from .rxdException import RxDException

# import sys
# if 'neuron.rxd' in sys.modules:
#    raise RxDException('NEURON CRxD module cannot be used with NEURON RxD module.')

from . import rxd, constants
from .species import Species, Parameter, State
from .region import Region, Extracellular
from .rate import Rate
from .reaction import Reaction
from . import geometry
from .multiCompartmentReaction import MultiCompartmentReaction
from .rxd import re_init, set_solve_type, nthread
from .rxdmath import v

try:
    from . import dimension3
except:
    pass
from .rangevar import RangeVar
from .geometry import (
    membrane,
    inside,
    Shell,
    FractionalVolume,
    FixedCrossSection,
    FixedPerimeter,
    ScalableBorder,
    DistributedBoundary,
    MultipleGeometry,
)
from .plugins import set_solver

# deprecated:
# from geometry import ConstantArea, ConstantVolume
# TODO: if we ever separate Parameter and State from species, then we need to
#       rembember to call rxd._do_nbs_register()


def _model_view(tree):
    from . import species
    from neuron import h

    species_dict = species._get_all_species()
    if "TreeViewItem" not in dir(h):
        return
    if species_dict:
        rxd_head = h.TreeViewItem(None, "Reaction Diffusion Items")
        # TODO: do the species disappear if they go out of scope? or does this overcount?
        rxd_species = h.TreeViewItem(
            rxd_head, "%d Species/State/Parameter" % len(species_dict)
        )
        species_children = [
            h.TreeViewItem(rxd_species, str(name)) for name in species_dict
        ]
        rxd_reactions = h.TreeViewItem(
            rxd_head,
            "%d Reaction/Rate/MultiCompartmentReaction"
            % len([r for r in rxd._all_reactions if r() is not None]),
        )
        tree.append(rxd_head)


def save_state():
    """return a bytestring representation of the current rxd state

    Note: this is dependent on the order items were created."""
    from . import species
    import array
    import itertools
    import gzip

    version = 0
    state = []
    num_species = 0
    for sp in species._all_species:
        s = sp()
        if s is not None:
            my_state = s._state
            state.append(array.array("Q", [len(my_state)]).tobytes())
            state.append(my_state)
            num_species += 1
    if num_species == 0:
        return b""
    data = gzip.compress(
        array.array("Q", [num_species]).tobytes()
        + bytes(itertools.chain.from_iterable(state))
    )
    metadata = array.array("Q", [version, len(data)]).tobytes()
    return metadata + data


def restore_state(oldstate):
    """restore rxd state from a bytestring

    Note: this is dependent on the order items were created."""
    from . import species
    import array
    import itertools
    import gzip

    if oldstate == b"":
        for sp in species._all_species:
            s = sp()
            if s is not None:
                raise RxDException("Invalid state data: inconsistent number of Species")
        return
    metadata = array.array("Q")
    metadata.frombytes(oldstate[:16])
    version, length = metadata
    if version != 0:
        raise RxDException("Unsupported state version")
    # discard header and decompress remainder
    if len(oldstate) != length + 16:
        raise RxDException("Invalid state data: bad length")
    oldstate = gzip.decompress(oldstate[16:])
    metadata = array.array("Q")
    metadata.frombytes(oldstate[:8])
    num_species = metadata[0]
    active_species = []
    for sp in species._all_species:
        s = sp()
        if s is not None:
            active_species.append(s)
    if len(active_species) != num_species:
        raise RxDException("Invalid state data: inconsistent number of Species")
    position = 8
    for sp in active_species:
        data = array.array("d")
        size_array = array.array("Q")
        size_array.frombytes(oldstate[position : position + 8])
        position += 8
        size = size_array[0]
        data.frombytes(oldstate[position : position + size])
        position += size
        sp._state = bytes(data)
    if position != len(oldstate):
        raise RxDException("Invalid state data: bad length")



import pytest
import numpy
from testutils import compare_data, tol


@pytest.fixture
def ics_diffusion_hybrid(neuron_instance):
    """A model using intracellular diffusion in a 1D and 3D sections"""

    h, rxd, data, save_path = neuron_instance
    dend1 = h.Section(name="dend1")
    dend1.diam = 2
    dend1.nseg = 1
    dend1.L = 10

    dend2 = h.Section(name="dend2")
    dend2.diam = 2
    dend2.nseg = 11
    dend2.L = 10

    dend3 = h.Section(name="dend3")
    dend3.diam = 2
    dend3.nseg = 1
    dend3.L = 10

    dend2.connect(dend1)
    dend3.connect(dend2)

    diff_constant = 1

    r = rxd.Region(h.allsec(), dx=0.75)

    rxd.set_solve_type([dend2], dimension=3)

    ca = rxd.Species(
        r,
        d=diff_constant,
        initial=lambda node: 1
        if (0.8 < node.x and node in dend1) or (node.x < 0.2 and node in dend2)
        else 0,
    )
    model = ([dend1, dend2, dend3], r, ca)
    yield (neuron_instance, model)


def test_pure_diffusion_hybrid(ics_diffusion_hybrid):
    """Test ics_diffusion_hybrid with fixed step methods"""

    neuron_instance, model = ics_diffusion_hybrid
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = model
    h.dt *= 50
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
        max_err = compare_data(data)
        assert max_err < tol


def test_pure_diffusion_hybrid_cvode(ics_diffusion_hybrid):
    """Test ics_diffusion_hybrid with variable step methods"""

    neuron_instance, model = ics_diffusion_hybrid
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = model
    h.CVode().active(True)
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
        max_err = compare_data(data)
        assert max_err < tol


def test_pure_diffusion_hybrid_small_grid(ics_diffusion_hybrid):
    """Test ics_diffusion_hybrid with fixed step methods where 1D sections are
    outside the 3D grid
    """

    neuron_instance, model = ics_diffusion_hybrid
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = model
    dend[1].diam = 0.75

    h.dt *= 50
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
        max_err = compare_data(data)
        assert max_err < tol



from neuronmusic cimport *

cdef class EventOutputPort:
    cdef cxx_EventOutputPort* cxx # hold a C++ instance which we're wrapping
    
    def __cinit__ (self):
        self.cxx = NULL

    def __dealloc__ (self):
        del_EventOutputPort (self.cxx)

    def gid2index (self, int gid, int gi):
        if self.cxx != NULL:
            self.cxx.gid2index (gid, gi)

cdef wrapEventOutputPort (cxx_EventOutputPort* port):
    cdef EventOutputPort port_ = EventOutputPort ()
    port_.cxx = port
    return port_


cdef class EventInputPort:
    cdef cxx_EventInputPort* cxx # hold a C++ instance which we're wrapping
    
    def __cinit__ (self):
        self.cxx = NULL

    def __dealloc__ (self):
        del_EventInputPort (self.cxx)

    def index2target (self, int gi, target):
        if self.cxx != NULL:
            return self.cxx.index2target (gi, target)
        else:
            return False

cdef wrapEventInputPort (cxx_EventInputPort* port):
    cdef EventInputPort port_ = EventInputPort ()
    port_.cxx = port
    return port_


def publishEventOutput (identifier):
    return wrapEventOutputPort (cxx_publishEventOutput (identifier.encode ('utf-8')))


def publishEventInput (identifier):
    return wrapEventInputPort (cxx_publishEventInput (identifier.encode ('utf-8')))


# Local Variables:
# mode: python
# End:



def test_savestate(neuron_instance):
    """Test rxd SaveState

    Ensures restoring to the right point and getting the right result later.

    Note: getting the right result later was a problem in a prior version of
    NEURON, so the continuation test is important. The issue was that somehow
    restoring from a SaveState broke the rxd connection to the integrator.
    """

    h, rxd, data, save_path = neuron_instance

    soma = h.Section(name="soma")
    soma.insert(h.hh)
    soma.nseg = 51
    cyt = rxd.Region(h.allsec(), name="cyt")
    c = rxd.Species(cyt, name="c", d=1, initial=lambda node: 1 if node.x < 0.5 else 0)
    c2 = rxd.Species(
        cyt, name="c2", d=0.6, initial=lambda node: 1 if node.x > 0.5 else 0
    )
    r = rxd.Rate(c, -c * (1 - c) * (0.3 - c))
    r2 = rxd.Reaction(c + c2 > c2, 1)

    h.finitialize(-65)
    soma(0).v = -30

    while h.t < 5:
        h.fadvance()

    def get_state():
        return (
            soma(0.5).v,
            c.nodes(soma(0.5)).concentration[0],
            c2.nodes(soma(0.5)).concentration[0],
        )

    s1 = get_state()
    s = h.SaveState()
    s.save()

    while h.t < 10:
        h.fadvance()

    s2 = get_state()
    s.restore()

    assert get_state() == s1
    assert get_state() != s2

    while h.t < 10:
        h.fadvance()

    assert get_state() != s1
    assert get_state() == s2
